import tempfile

from fabric.api import put, sudo

from .service import Service
from ..hooks import hook
from ..shared import apt_get_install, pip_install, chown

class Memcached(Service):
	name = 'memcached'
	script = 'memcached'
	DEFAULT_SETTINGS = {'-d': None, 'logfile': '/var/log/memcached.log', '-m': '64', '-p': '11211', '-u': 'memcache', '-l': '127.0.0.1'}

	def install(self):
		with hook('install %s' % self.name, self):
			apt_get_install('memcached')

	def config(self):
		settings = self.settings if self.settings else Memcached.DEFAULT_SETTINGS
		with hook('config %s' % self.name, self):
			with tempfile.NamedTemporaryFile('w') as f:
				for setting in settings:
					if settings[setting]:
						f.write('%s %s\n' % (setting, settings[setting]))
					else:
						f.write('%s\n' % setting)
				f.flush()
				chown(put(f.name, '/etc/memcached.conf', use_sudo=True, mode=0644))
		self.restart()

	def site_install(self, site):
		with hook('site install %s' % self.name, self, site):
			pip_install(site, 'python-memcached')



import os
import shutil
import tempfile

from django.conf import settings
from django.core.urlresolvers import resolve
from django.test.client import RequestFactory

def abstract_model_copy(model, exclude=None):
	"""Returns a copy of a model to be used as an abstract base class."""
	attrs = {}
	for field in model._meta.fields:
		if field.name != 'id' and field.name not in exclude:
			attrs[field.name] = field
	attrs['__module__'] = model.__module__
	attrs['Meta'] = type('Meta', (), {'abstract':True})
	copy = type(model.__name__ + 'Copy', model.__bases__, attrs)
	return copy

def have_fields_changed(obj, *fields):
	"""Returns True if all the specified fields on an object have changed compared the current database version."""
	if not obj.pk:
		return False
	try:
		current_obj = obj.__class__.objects.only(*fields).get(pk=obj.pk)
		for field in fields:
			if getattr(obj, field) == getattr(current_obj, field):
				return False
		return True
	except:
		return False

def __publish_object(obj, unpublish=False, callback=None):
	if not hasattr(obj, 'get_absolute_url'):
		raise ValueError('%s is not setup for publishing' % repr(obj))
	if hasattr(obj.__class__, 'Publishing'):
		publishing = obj.__class__.Publishing
	else:
		publishing = None
	# Configure the publish and unpublish methods
	slug_field = getattr(publishing, 'slug_field', 'slug')
	slug = getattr(obj, slug_field, str(obj.pk))
	prefix = getattr(publishing, 'prefix', '')
	if not prefix.endswith('/'):
		prefix = '%s/%s/' % (prefix.lstrip('/'), slug)
	else:
		prefix = '%s%s/' % (prefix.lstrip('/'), slug)
	kwargs = {'callback': callback, 'prefix': prefix}
	if hasattr(settings, 'AWS_ACCESS_KEY_ID') and hasattr(settings, 'AWS_SECRET_ACCESS_KEY'):
		from aws import upload_to_s3, delete_from_s3
		upload = upload_to_s3
		delete = delete_from_s3
		kwargs['bucket'] = getattr(publishing, 'bucket', settings.AWS_STORAGE_BUCKET_NAME)
	else:
		raise Exception('No publishing data store is available')
		
	if unpublish:
		# Unpublishing (deleting)
		delete(**kwargs)
		obj.update(published=None)
	else:
		# Publishing
		tempdir = tempfile.mkdtemp()
		rf = RequestFactory()
		url = obj.get_absolute_url()
		request = rf.get(url)
		view, view_args, view_kwargs = resolve(url)
		response = view(request, *view_args, **view_kwargs)
		# If the url has no extension, the convert it to index.html
		if not len(os.path.splitext(url)[1]):
			filename = '%s/index.html' % tempdir
		else:
			filename = os.path.join(tempdir, url.strip('/'))
		with open(filename, 'w') as f:
			f.write(response.content if not response.streaming else b''.join(response.streaming_content))
		subpages = getattr(publishing, 'subpages', [])
		for subpage in subpages:
			url = os.path.join(obj.get_absolute_url(), subpage)
			request = rf.get(url)
			view, view_args, view_kwargs = resolve(url)
			response = view(request, *view_args, **view_kwargs)
			if not len(os.path.splitext(subpage)[1]):
				filename = '%s/index.html' % os.path.join(tempdir, subpage.strip('/'))
			else:
				filename = os.path.join(tempdir, subpage.strip('/'))
			try:
				os.makedirs(os.path.split(filename)[0])
			except:
				pass
			with open(filename, 'w') as f:
				f.write(response.content if not response.streaming else b''.join(response.streaming_content))
		upload(directory=tempdir, **kwargs)
		shutil.rmtree(tempdir)

def publish_object(obj, callback=None):
	"""Publish or delete an object onto a production data store.

	An object's get_absolute_url method MUST be implemented. index.html will be appended to any url ending with /
	NOTE: When publishing from a development command be sure to sync media files and the object itself (sync_data and sync_media) and AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY need to be in the development settings.
		model.Publishing values:
			prefix - the prefix to use on all uploaded files, default is ''
			slug_field - the obj attribute to use as the unique path under prefix, default is slug
			subpages - list of subpages to render and publish: urls relative to this object's get_absolute_url() e.g. subpage/
			bucket - for S3 only, the bucket to use
	"""
	__publish_object(obj, callback=callback)

def unpublish_object(obj, callback=None):
	__publish_object(obj, True, callback=callback)

def sync_objects(queryset, other_database='default'):
	"""Sync data between one database and another from a queryset. Will also sync any select_related data."""
	def __sync_related(obj, related_dict):
		# Recursively save the related objects
		for field in related_dict:
			related_obj = getattr(obj, field)
			related_obj.save(using=other_database)
			if related_dict[field]:
				__sync_related(related_obj, related_dict[field])

	for obj in queryset:
		# query.select_related will equal either False or e.g. {'field': {'subfield': {}}} for select_related('field__subfield')
		if queryset.query.select_related:
			__sync_related(obj, queryset.query.select_related)
		obj.save(using=other_database)

# Media methods should be used within management commands only

def __load_production_default_storage(production_settings):
	production_settings_module = __import__(production_settings, globals(), locals(), ['DEFAULT_FILE_STORAGE'], -1)
	file_storage_class = production_settings_module.DEFAULT_FILE_STORAGE
	file_storage_class = file_storage_class.split('.')
	file_storage_module = __import__('.'.join(file_storage_class[:-1]), globals(), locals(), [file_storage_class[-1]], -1)
	file_storage_class = getattr(file_storage_module, file_storage_class[-1])
	file_storage = file_storage_class()
	return file_storage

def __sync_media(media_files, src_storage, dest_storage, ignore_missing, callback):
	"""Saves media from src_storage into dest_storage."""
	for media_file in media_files:
		if not ignore_missing and not src_storage.exists(media_file):
			raise IOError('No such file %s' % media_file)
		elif callback:
			callback(media_file)
		f = src_storage.open(media_file)
		dest_storage.delete(media_file)
		dest_storage.save(media_file, f)
		f.close()

def upload_media(media_files, production_settings, ignore_missing=True, callback=None):
	"""
	Upload a subset of media files to the production data store. Both the development and production DEFAULT_FILE_STORAGE classes are used.
	Appropriate access keys etc. will need to also be defined in development settings.
	"""
	if media_files:
		from django.core.files.storage import default_storage
		production_default_storage = __load_production_default_storage(production_settings)
		__sync_media(media_files, default_storage, production_default_storage, ignore_missing, callback)

def download_media(media_files, production_settings, ignore_missing=True, callback=None):
	"""
	Download a subset of media files from a production data store. Both the development and production DEFAULT_FILE_STORAGE classes are used.
	Appropriate access keys etc. will need to also be defined in development settings.
	"""
	if media_files:
		from django.core.files.storage import default_storage
		production_default_storage = __load_production_default_storage(production_settings)
		__sync_media(media_files, production_default_storage, default_storage, ignore_missing, callback)



from django.conf import settings
from django.db import connection

"""
http://dev.mysql.com/doc/refman/5.0/en/lock-tables-and-transactions.html
In mysql locks cannot be used with transactions, do not use this 

Use this as a last resort.  Use select_for_update or just update if possible first.
This does not play well with the transaction Django sets up with it's middleware and with testing.
NOTE: This function needs something to effectively end the current transaction and then start a new one to deal with the Transaction middleware. See transaction middleware for ideas.
"""

# Generate a lock decorator using a list of models to read lock and a list to write lock
def lock_models(read_lock_models, write_lock_models):
	"""Lock tables based on a list of models to read or write lock.

	Read locks mean that no one including the current session can write to the table.
	Write locks mean that no one except the current session can write to the table, but anyone can read.
	Pass the arguments a lists or None e.g. lock_models((mymodel1,mymodel2), (mymodel3,)) or lock_models(None, (mymodel3,))
	"""
	def lock_decorator(func):
		def lock_fun(*args, **kwargs):
		
			if settings.DATABASES['default']['ENGINE'].find('mysql') != -1:
				if read_lock_models:
					read_lock_str = ", ".join(map(lambda model:"%s READ" % (model._meta.db_table,), read_lock_models))
				else:
					read_lock_str = None
				if write_lock_models:
					write_lock_str = ", ".join(map(lambda model:"%s WRITE" % (model._meta.db_table,), write_lock_models))
				else:
					write_lock_str = None
					
				if read_lock_str and write_lock_str:
					lock_str = ", ".join((read_lock_str, write_lock_str))
				elif read_lock_str:
					lock_str = read_lock_str
				elif write_lock_str:
					lock_str = write_lock_str
				else:
					return func(*args, **kwargs)
				
				cursor = connection.cursor()
				# Lock
				cursor.execute("SET autocommit=0;")
				cursor.execute("LOCK TABLES %s;" % (lock_str,))
				# Work
				try:
					result = func(*args, **kwargs)
				except Exception as e:
					cursor.execute("ROLLBACK;UNLOCK TABLES;")
					raise e
				# Unlock
				cursor.execute("COMMIT;UNLOCK TABLES;")
				return result
				
			elif settings.DATABASES['default']['ENGINE'].find('postgres') != -1:
				if not read_lock_models and not write_lock_models:
					return func(*args, **kwargs)
				cursor = connection.cursor()
				# Lock
				cursor.execute("BEGIN;")
				if read_lock_models:
					cursor.execute("LOCK TABLE %s IN ACCESS SHARE;" % (",".join(map(lambda model:model._meta.db_table, read_lock_models)),))
				if write_lock_models:
					cursor.execute("LOCK TABLE %s IN EXCLUSIVE;" % (",".join(map(lambda model:model._meta.db_table, write_lock_models)),))
				# Work
				try:
					result = func(*args, **kwargs)
				except Exception as e:
					cursor.execute("ROLLBACK;")
					raise e
				# Unlock
				cursor.execute("COMMIT;")
				
			else:
				return func(*args, **kwargs)
			
		return lock_fun
	return lock_decorator



from django import template

register = template.Library()

@register.filter(is_safe=True)
def negativeparens(value):
	"""Surrounds the value with parentheses if it starts with a -."""
	value = str(value)
	if value[0] == '-':
		return '(%s)' % value[1:]
	return value



import StringIO

from fabric.api import run, sudo, cd, put, env, settings, hide
from fabric.context_managers import shell_env
from fabric.contrib.files import append, exists

from .postgis import *
from .service import Service
from ..hooks import hook
from ..shared import apt_get_install, pip_install, find_service, chown, get_django_setting, SiteType

POSTGRES_USER = 'postgres'
POSTGRES_GROUP = 'postgres'

_SHOW_FILE_COMMAND = '$(sudo -u %s psql -t -P format=unaligned -c "SHOW %%s;")' % POSTGRES_USER
_CONFIG_FILE_PATH = _SHOW_FILE_COMMAND % 'config_file'
_HBA_FILE_PATH = _SHOW_FILE_COMMAND % 'hba_file'
_IDENT_FILE_PATH = _SHOW_FILE_COMMAND % 'ident_file'
_MAIN_DIR = '$(dirname %s)' % _CONFIG_FILE_PATH
_CONF_DIR = '%s/conf.d' % _MAIN_DIR
_EXCLUDE_SETTINGS = ['postgis', 'hba', 'ident']
# NOTE: If hba is set to True instead of a list, then client authentication for the current local host is added

def _pg_quote_config(key, value):
	if (isinstance(value, (str, unicode)) and value not in ('on', 'off') and not value[0].isdigit()) or key == 'listen_addresses':
		return "'%s'" % value
	return value

def _get_pg_env(site):
	database = get_django_setting(site, 'DATABASES')['default']
	return {
		'PGHOST': database['HOST'],
		'PGPORT': str(database.get('PORT') or '5432'),
		'PGUSER': database['USER'],
		'PGPASSWORD': database['PASSWORD'],
		'PGDATABASE': database['NAME']
	}

class Postgres(Service):
	name = 'postgres'
	script = 'postgresql'

	def execute_sql(self, raw_sql, site=None):
		sql = []
		for line in raw_sql.split('\n'):
			line = line.strip()
			if not line or line.startswith('--'):
				continue
			sql.append(line.replace('\t', '').replace('\n', '').replace("'", "\\'"))
		sql = ' '.join(sql)
		if site:
			with shell_env(**_get_pg_env(site)), settings(warn_only=True):
				sudo("psql --echo-queries -c $'%s'" % sql)
		else:
			with settings(warn_only=True):
				sudo("psql --echo-queries -c $'%s'" % sql, user=POSTGRES_USER)

	def execute_command(self, pg_command, site=None):
		if site:
			with shell_env(**_get_pg_env(site)), settings(warn_only=True):
				sudo(pg_command)
		else:
			with settings(warn_only=True):
				sudo(pg_command, user=POSTGRES_USER)

	def install(self):
		with hook('install %s' % self.name, self):
			apt_get_install('postgresql')
			if self.settings.get('postgis'):
				install_postgis()

	def config(self):
		with hook('config %s' % self.name, self):
			if not exists(_CONF_DIR):
				sudo('mkdir %s' % _CONF_DIR)
				chown(_CONF_DIR, POSTGRES_USER, POSTGRES_GROUP)
				append(_CONFIG_FILE_PATH, "include_dir 'conf.d'", use_sudo=True)
			if self.settings:
				# Apply any given settings and place into a new conf.d directory
				config = ''
				if env.get('vagrant'):
					self.settings['listen_addresses'] = '*'
				for setting in self.settings:
					if setting not in _EXCLUDE_SETTINGS:
						config += '%s = %s\n' % (setting, _pg_quote_config(setting, self.settings[setting]))
				if config:
					chown(put(StringIO.StringIO(config), _CONF_DIR + '/fabric.conf', use_sudo=True, mode=0644), POSTGRES_USER, POSTGRES_GROUP)
				# Apply any given Client Authentications given under hba
				hba = list(self.settings.get('hba', []))
				if env.get('vagrant') or hba == True:
					if hba == True:
						hba = []
					with hide('everything'):
						host_ip = run('echo $SSH_CLIENT').split(' ')[0]
					hba.append(('host', 'all', 'all', host_ip + '/32', 'md5'))
				if hba:
					append(_HBA_FILE_PATH, '# Fabric client connections:', use_sudo=True)
				for client in hba:
					client = '%s%s%s%s%s' % (client[0].ljust(8), client[1].ljust(16), client[2].ljust(16), client[3].ljust(24), client[4])
					append(_HBA_FILE_PATH, client, use_sudo=True)
				# Apply any given identity mappings
				ident = self.settings.get('ident', [])
				if ident:
					append(_IDENT_FILE_PATH, '# Fabric username maps:', use_sudo=True)
				for mapping in ident:
					mapping = '%s%s%s' % (mapping[0].ljust(16), mapping[1].ljust(24), mapping[2])
					append(_IDENT_FILE_PATH, mapping, use_sudo=True)
		self.restart()

	def site_install(self, site):
		with hook('site install %s' % self.name, self, site):
			if self.settings.get('postgis'):
				# Install PostGIS also on the site if separate from the server
				if find_service(self.name) is None:
					install_postgis()
			# Install python postgresql support
			pip_install(site, 'psycopg2')

	def site_config(self, site):
		with hook('site config %s' % self.name, self, site):
			if site['type'] == SiteType.DJANGO:
				database = get_django_setting(site, 'DATABASES')['default']
				# Create the user for django to access the database with
				if find_service(self.name) is not None:
					with settings(warn_only=True):
						self.execute_command('createuser --echo --createdb --no-superuser --no-createrole ' + database['USER'])
						self.execute_sql("ALTER USER %s WITH PASSWORD '%s';" % (database['USER'], database['PASSWORD']))
				# Create the database
				self.execute_command('createdb --echo ' + database['NAME'], site)
			# Setup postgis
			if self.settings.get('postgis'):
				site_config_postgis(self, site)




# Works on 0-based index, with 1-based index text
def pagination_pages(current_page, total_pages, pagination_size=5, space_text='...', prev_text='<', next_text='>'):
	class PaginationPage:
		def __init__(self, enabled, text, number=1):
			self.enabled = enabled
			self.text = text
			self.number = number
	pages = []
	pagination_size /= 2
	lower = current_page - pagination_size
	upper = current_page + pagination_size
	if lower < 0:
		upper += -lower
		lower = 0
	if upper >= total_pages:
		upper = total_pages - 1
	# Back button
	pages.append(PaginationPage(True if current_page > 0 else False, prev_text, current_page - 1))
	# Spacer
	if lower > 0:
		pages.append(PaginationPage(False, space_text))
	# Number buttons
	for x in xrange(lower, upper + 1):
		pages.append(PaginationPage(False if x == current_page else True, str(x + 1), x))
	# Spacer
	if upper < (total_pages - 1):
		pages.append(PaginationPage(False, space_text))
	# Next button
	pages.append(PaginationPage(True if current_page < (total_pages - 1) else False, next_text, current_page + 1))
	return pages



from fabric.api import sudo, env
from fabric.contrib.files import append

from .nginx import NGINX_USER
from .service import Service
from ..hooks import hook

class Cron(Service):
	name = 'cron'
	script = None

	def site_config(self, site):
		# Creates a new crontab for the nginx user.
		# list of (5,6,'*',0,'full shell command') or just 'management command'
		# http://www.adminschoice.com/crontab-quick-reference/
		with hook('site config %s' % self.name, self, site):
			# Get the existing crontab
			result = sudo('crontab -l -u %s' % NGINX_USER, warn_only=True)
			lines = result.splitlines() if result.succeeded else []
			start = '# start [fabric] [%s]' % site['name']
			end = '# end [fabric] [%s]' % site['name']
			try:
				# Remove the previous section
				start_index = lines.index(start)
				end_index = lines.index(end)
				lines = lines[0:start_index] + lines[end_index+1:]
			except:
				pass
			if self.settings.get('crontab'):
				lines.append(start)
				for job in self.settings['crontab']:
					if isinstance(job, (str, unicode)):
						lines.append('0 0 * * * cd %s && python manage.py %s --settings %s >/dev/null 2>&1' % ('/srv/www/%s' % env['project'], job, site['settings_module']))
					else:
						lines.append(' '.join())
				lines.append(end)
			if lines:
				import sha
				lines = '\n'.join(lines)
				crontab_file = '/tmp/fabric_crontab_' + sha.new(lines).digest().encode('hex')[0:10]
				sudo('touch ' + crontab_file)
				append(crontab_file, lines, use_sudo=True)
				sudo('crontab -u %s %s' % (NGINX_USER, crontab_file))
				sudo('rm ' + crontab_file)
			else:
				sudo('crontab -r %s' % NGINX_USER, warn_only=True)



import StringIO

from fabric.api import sudo, put
from fabric.contrib.files import append

from .service import Service
from ..hooks import hook
from ..shared import WWW_USER, get_python_interpreter, get_project_directory, SiteType, chown

class CronSchedule(object):
	def __init__(self, minute='0', hour='0', day_of_month='*', month='*', day_of_week='*'):
		self.minute = str(minute)
		self.hour = str(hour)
		self.day_of_month = str(day_of_month)
		self.month = str(month)
		self.day_of_week = str(day_of_week)

	def __str__(self):
		return ' '.join((self.minute, self.hour, self.day_of_month, self.month, self.day_of_week))

class CronJob(object):
	def __init__(self, command=None, **kwargs):
		self.schedule = CronSchedule()
		self.command = command
		self.management_command = None
		self.directory = None
		self.log_name = None
		self.log_rotate = ('compress', 'daily', 'missingok', 'rotate 30')
		self.chain = None
		self.__dict__.update(kwargs)
		# Remove the schedule from chain because they are on the same
		if self.chain:
			self.chain.schedule = None

	def get_cron_entry(self, site):
		components = []
		if self.schedule:
			components.append(str(self.schedule))
		if self.management_command:
			if site['type'] == SiteType.DJANGO:
				components.append('cd %s &&' % get_project_directory())
				components.append('%s manage.py %s --settings %s' % (get_python_interpreter(site), self.management_command, site['settings_module']))
		elif self.command:
			if self.directory:
				components.append('cd %s &&' % self.directory)
			components.append(self.command)
		if self.log_name:
			components.append('>>/var/log/%s/%s.log 2>&1' % (self.log_name, self.log_name))
		else:
			components.append('>/dev/null 2>&1')
		if self.chain:
			components.append('&&')
			components.append(str(self.chain.get_cron_entry(site)))
		return ' '.join(components)

class ManagementJob(CronJob):
	def __init__(self, management_command, **kwargs):
		kwargs['management_command'] = management_command
		super(ManagementJob, self).__init__(**kwargs)

def read_crontab(user):
	result = sudo('crontab -u %s -l' % user, warn_only=True)
	return result.splitlines() if result.succeeded else []

def write_crontab(user, lines):
	if lines:
		import sha
		lines = '\n'.join(lines)
		crontab_file = '/tmp/fabric_crontab_' + sha.new(lines).digest().encode('hex')[0:10]
		sudo('touch ' + crontab_file)
		append(crontab_file, lines, use_sudo=True)
		sudo('crontab -u %s %s' % (user, crontab_file))
		sudo('rm ' + crontab_file)
	else:
		sudo('crontab -u %s -r' % user, warn_only=True)

def add_crontab_section(user, section_name, jobs, site):
	if not jobs:
		return
	lines = read_crontab(user)
	start = '# start %s' % section_name
	end = '# end %s' % section_name
	lines.append(start)
	if isinstance(jobs, CronJob):
		jobs = [jobs]
	for job in jobs:
		lines.append(job.get_cron_entry(site))
	lines.append(end)
	write_crontab(user, lines)

def remove_crontab_section(user, section_name):
	lines = read_crontab(user)
	start = '# start %s' % section_name
	end = '# end %s' % section_name
	try:
		# Remove the previous section
		start_index = lines.index(start)
		end_index = lines.index(end)
		lines = lines[0:start_index] + lines[end_index+1:]
	except:
		pass
	write_crontab(user, lines)

class Cron(Service):
	name = 'cron'
	script = None

	def site_config(self, site):
		# Creates a new crontab for the nginx user.
		# crontab is a single or list of: CronJobs
		# http://www.adminschoice.com/crontab-quick-reference/
		with hook('site config %s' % self.name, self, site):
			section_name = '[fabric] [%s]' % site['name']
			jobs = self.settings.get('crontab')
			if isinstance(jobs, CronJob):
				jobs = [jobs]
			# Remove then add the existing crontab section
			remove_crontab_section(WWW_USER, section_name)
			add_crontab_section(WWW_USER, section_name, jobs, site)
			# Setup logging directories and logrotate.d
			for job in jobs:
				if job.log_name:
					sudo('mkdir -p /var/log/%s' % job.log_name)
					if job.log_rotate:
						rotate_conf = StringIO.StringIO('/var/log/%s/%s.log {\n    %s\n}\n' % (job.log_name, job.log_name, '\n    '.join(job.log_rotate)))
						chown(put(rotate_conf, '/etc/logrotate.d/%s' % job.log_name, use_sudo=True, mode=0644))



from django import template
from django.conf import settings

register = template.Library()

@register.filter
def absoluteurl(path, domain=None):
	"""Builds an absolute url with either the provided domain (typically taken from get_current_site) or the first domain found in settings.ALLOWED_HOSTS."""
	if path and type(path) != unicode:
		path = unicode(path)
	if path and not path.startswith('http://'):
		if domain:
			if path.startswith('/'):
				return 'http://%s%s' % (domain, path)
			else:
				return 'http://%s/%s' % (domain, path)
		elif len(settings.ALLOWED_HOSTS):
			domain = settings.ALLOWED_HOSTS[0]
			if domain.startswith('.'):
				domain = domain[1:]
			if path.startswith('/'):
				return 'http://%s%s' % (domain, path)
			else:
				return 'http://%s/%s' % (domain, path)
	return path



import urlparse

from django.http import HttpResponse

def enable_cors(origins, methods, allow_credentials):
	"""Enable Cross-origin resource sharing on a view, by responding to pre-flight requests and adding appropriate headers.

	origins - a list of allowed hostnames, or empty or None for all hostnames
	methods - a list of allowed methods in caps, or copy the requests method if None
	allow_credentials - allow the client to pass cookies

	http://www.html5rocks.com/en/tutorials/cors/
	http://www.w3.org/TR/cors/
	https://developer.mozilla.org/en-US/docs/Web/HTTP/Access_control_CORS
	"""
	def enable_cors_decorator(func):
		def enable_cors_fun(request, *args, **kwargs):
			# Only handle if the Origin header is present
			if not request.META.has_key('HTTP_ORIGIN'):
				return func(request, *args, **kwargs)
			if request.method == 'OPTIONS':
				if origins and urlparse.urlparse(request.META['HTTP_ORIGIN']).netloc not in origins:
					return HttpResponse(status=403)
				if methods and request.META.get('HTTP_ACCESS_CONTROL_REQUEST_METHOD', 'GET') not in methods:
					return HttpResponse(status=405)
				response = HttpResponse(status=204)
			else:
				response = func(request, *args, **kwargs)
			# Copy the request's Origin
			response['Access-Control-Allow-Origin'] = request.META['HTTP_ORIGIN']
			# Set the allowed methods
			if methods:
				response['Access-Control-Allow-Methods'] = ', '.join(methods)
			else:
				response['Access-Control-Allow-Methods'] = request.META.get('HTTP_ACCESS_CONTROL_REQUEST_METHOD', 'GET, POST, UPDATE, DELETE')
			# Copy the request's Access-Control-Request-Headers
			if request.META.has_key('HTTP_ACCESS_CONTROL_REQUEST_HEADERS'):
				response['Access-Control-Allow-Headers'] = request.META['HTTP_ACCESS_CONTROL_REQUEST_HEADERS']
			if allow_credentials:
				response['Access-Control-Allow-Credentials'] = 'true'
			return response
		return enable_cors_fun
	return enable_cors_decorator



import os
import select
import socket
import SocketServer
import thread

from django.conf import settings
from django.db import connections
from django.db.utils import ConnectionDoesNotExist

"""
Tunneling mostly based on the sample paramiko code:
https://github.com/paramiko/paramiko/blob/master/demos/forward.py
Previously tried unsuccessfully based on using the subprocess module and runnning:
args = ['ssh', '-f', '-N', '-o', 'StrictHostKeyChecking=no', '-o', 'ExitOnForwardFailure=yes', '-L']
in addition to specifying the host, username, and identityfile
"""

SSH_CLIENT_KEY = '__DATABASE_SSH_CLIENT'
SSH_TUNNEL_KEY = '__DATABASE_SSH_TUNNEL'

class __PortForwardingServer(SocketServer.ThreadingTCPServer):
	daemon_threads = True
	allow_reuse_address = True

class __PortForwardingServerHandler(SocketServer.BaseRequestHandler):
	def handle(self):
		try:
			chan = self.ssh_transport.open_channel('direct-tcpip', (self.chain_host, self.chain_port), self.request.getpeername())
		except Exception, e:
			#print 'Incoming request to %s:%d failed: %s' % (self.chain_host, self.chain_port, repr(e))
			return
		if chan is None:
			#print('Incoming request to %s:%d was rejected by the SSH server.' % (self.chain_host, self.chain_port))
			return
		#print('Connected!  Tunnel open %r -> %r -> %r' % (self.request.getpeername(), chan.getpeername(), (self.chain_host, self.chain_port)))
		while True:
			r, w, x = select.select([self.request, chan], [], [])
			if self.request in r:
				data = self.request.recv(1024)
				if len(data) == 0:
					break
				chan.send(data)
			if chan in r:
				data = chan.recv(1024)
				if len(data) == 0:
					break
				self.request.send(data)
		chan.close()
		self.request.close()
		#print('Tunnel closed from %r' % (self.request.getpeername(),))

def __start_tunnel(server):
	server.serve_forever()

def start_tunnel(database, use_ssh_config=False):
	if not database:
		return
	from paramiko import AutoAddPolicy, SSHClient, SSHConfig
	db = settings.DATABASES[database]
	if db.has_key(SSH_CLIENT_KEY):
		# Tunnel is already running
		return
	if not db.has_key('REMOTE_HOST'):
		raise ValueError('REMOTE_HOST not specified for ' + database)
	if not db.has_key('TUNNEL_HOST'):
		raise ValueError('TUNNEL_HOST not specified for ' + database)
	kwargs = {}
	hostname = db['TUNNEL_HOST']
	
	# Setup the kwargs
	if db.has_key('TUNNEL_USER'):
		kwargs['username'] = db['TUNNEL_USER']
	if db.has_key('TUNNEL_PASSWORD'):
		kwargs['password'] = db['TUNNEL_PASSWORD']
	if db.has_key('TUNNEL_IDENTITY'):
		kwargs['key_filename'] = db['TUNNEL_IDENTITY']
	if db.has_key('TUNNEL_PORT'):
		kwargs['port'] = int(db['TUNNEL_PORT'])
	if use_ssh_config:
		try:
			with open(os.path.expanduser('~/.ssh/config')) as f:
				sshConfig = SSHConfig()
				sshConfig.parse(f)
				config = sshConfig.lookup(db['TUNNEL_HOST'])
				hostname = config['hostname']
				# Use username and port if missing
				if not kwargs.has_key('username') and config.has_key('user'):
					kwargs['username'] = config['user']
				if not kwargs.has_key('port') and config.has_key('port'):
					kwargs['port'] = int(config['port'])
				# Add identityfile (a list)
				if config.has_key('identityfile'):
					if kwargs.has_key('key_filename'):
						if type(kwargs['key_filename']) is list:
							kwargs['key_filename'] += config['identityfile']
						else:
							kwargs['key_filename'] = [kwargs['key_filename']] + config['identityfile']
					else:
						kwargs['key_filename'] = config['identityfile']
		except:
			pass

	# Fix the identity files
	if kwargs.has_key('key_filename'):
		if type(kwargs['key_filename']) is list:
			for i in range(len(kwargs['key_filename'])):
				if kwargs['key_filename'][i].startswith('~'):
					kwargs['key_filename'][i] = os.path.expanduser(kwargs['key_filename'][i])
		elif kwargs['key_filename'].startswith('~'):
			kwargs['key_filename'] = os.path.expanduser(kwargs['key_filename'])

	# Setup the client
	client = SSHClient()
	client.load_system_host_keys()
	client.set_missing_host_key_policy(AutoAddPolicy())
	client.connect(hostname, **kwargs)

	# Setup the port forwarding server
	class __SubPortForwardingServerHandler(__PortForwardingServerHandler):
		chain_host = db['REMOTE_HOST']
		chain_port = int(db['PORT'])
		ssh_transport = client.get_transport()
	server = __PortForwardingServer(('', int(db['PORT'])), __SubPortForwardingServerHandler)
	# Save a reference to the client and port forwarding server
	db[SSH_TUNNEL_KEY] = server
	db[SSH_CLIENT_KEY] = client
	# Start port forwarding server on another thread
	thread.start_new_thread(__start_tunnel, (server,))

def stop_tunnel(database):
	if not database:
		return
	# Close the database connection, because it will no longer be reachable after the tunnel goes down
	try:
		connections[database].close()
	except ConnectionDoesNotExist:
		pass
	db = settings.DATABASES[database]
	# Stop the server
	if db.has_key(SSH_TUNNEL_KEY):
		db[SSH_TUNNEL_KEY].shutdown()
		del db[SSH_TUNNEL_KEY]
	# Stop the client
	if db.has_key(SSH_CLIENT_KEY):
		db[SSH_CLIENT_KEY].close()
		del db[SSH_CLIENT_KEY]

class database_tunnel(object):
	"""Creates a temporary tunnel to a remote database for management commands or whatever else.

	Passing None will not raise any exceptions.
	Database dictionary in settings must include REMOTE_HOST and TUNNEL_HOST (optionally TUNNEL_PORT)
	Also, TUNNEL_USER and TUNNEL_PASSWORD or TUNNEL_IDENTITY must be specified, or have the TUNNEL_HOST setup in ~/.ssh/config.
	After setting up, make queries like: Model.objects.using('other').all()
	"""
	def __init__(self, database, use_ssh_config=False):
		self.__database = database
		self.__use_ssh_config = use_ssh_config
	def __enter__(self):
		start_tunnel(self.__database, self.__use_ssh_config)
	def __exit__(self, *_):
		stop_tunnel(self.__database)



from django.core.exceptions import ObjectDoesNotExist
from django.db.models import F
from django.db.models.fields import FieldDoesNotExist
from django.db.models.loading import cache

# These methods are most useful for logging and updating a stats model about an instance

def update_fields(instance, updates, select=False):
	"""Tries to update fields for an instance.

	Instance will not be updated upon return reflecting the updates!
	pre_ and post_save signals will not be sent when using this method!
	Use a normal select_for_update().get(), save() if the above is needed.
	select=True will give you pre_ post_save signals, but instance will not be updated.
	Only use update within an app with private fields and no signals are needed.
	"""
	if instance is None:
		return
	model = instance.__class__
	
	if select:
		try:
			select_instance = model.object.select_for_update().get(id=instance.id)
			for field_name in updates.keys():
				if hasattr(select_instance, field_name):
					setattr(select_instance, field_name, getattr(select_instance, field_name) + updates[field_name])
			select_instance.save()
		except ObjectDoesNotExist:
			pass
	else:
		# Do an actual update by resolving the updates into fields
		update_args = {}
		for field_name in updates.keys():
			try:
				field = model._meta.get_field_by_name(field_name)
				update_args[field_name] = F(field_name) + updates[field_name]
			except FieldDoesNotExist:
				pass
	
		if len(update_args):
			model.objects.filter(**{'id': instance.id}).update(**update_args)

def update_related_fields(model, related_field, related_instance, updates, select=False):
	"""Tries to update fields for a related instance.

	Take the same precautions as noted for the above update_fields.
	Optionally, pass select=True to use select_for_update and get the expected pre_ post_save signals.
	"""
	if model is None or related_instance is None:
		return
	if type(model) is str:
		modelStr = model.split('.')
		model = cache.get_model(modelStr[0], modelStr[1])
		
	if select:
		# Do the select_for_update
		try:
			instance = model.object.select_for_update().get(**{related_field: related_instance})
			for field_name in updates.keys():
				if hasattr(instance, field_name):
					setattr(instance, field_name, getattr(instance, field_name) + updates[field_name])
			instance.save()
		except ObjectDoesNotExist:
			pass
	else:
		# Do an actual update by resolving the updates into fields
		update_args = {}
		for field_name in updates.keys():
			try:
				field = model._meta.get_field_by_name(field_name)
				update_args[field_name] = F(field_name) + updates[field_name]
			except FieldDoesNotExist:
				pass
	
		if len(update_args):
			model.objects.filter(**{related_field: related_instance}).update(**update_args)



from importlib import import_module
import os
import subprocess
import sys
import time

from fabric.api import env, sudo, run, hide, settings, local
from fabric.contrib.files import exists

WWW_USER = 'www-data'

# Dictionary of environments set from the fab file.
environments = {}

class SiteType:
	DJANGO = 1
	WEBAPP = 2
	NODE = 3
	SCGI = 4
	NGINX = 5

presets = {
	'GZIP_CONTENT_TYPES': ('text/html', 'text/css','application/javascript','application/x-javascript', 'image/svg+xml'),
	'NO_CACHE': 'no-cache',
	'PERMANENT_CACHE': 'max-age=315360000'
}

# fabric.colors doesn't have bold, so just define styles here
def bold(msg):
	return '\033[1m%s\033[0m' % msg

def red(msg):
	return '\033[1;31m%s\033[0m' % msg

def green(msg):
	return '\033[1;32m%s\033[0m' % msg

def blue(msg):
	return '\033[1;34m%s\033[0m' % msg

def teal(msg):
	return '\033[1;36m%s\033[0m' % msg

class cd_local(object):
	"""Changes the local current working directory."""
	def __init__(self, nwd):
		self.cwd = os.getcwd()
		self.nwd = nwd
	def __enter__(self):
		os.chdir(self.nwd)
	def __exit__(self, *_):
		os.chdir(self.cwd)

def get_template_directory():
	return '%s/templates' % os.path.dirname(__file__)

def get_template(name):
	return os.path.join(get_template_directory(), name)

def chown(paths, username='root', group='root'):
	"""Change the owner and group of a path or multiple paths. The return value from put() or upload_template() can be wrapped in this directly."""
	if not paths:
		return
	if isinstance(paths, (str, unicode)):
		paths = [paths]
	for path in paths:
		sudo('chown -R %s:%s %s' % (username, group, path))

def fix_absolute_path(path):
	"""If settings spill over from development for a path inside the project then translate it to the project directory on the server."""
	if path.startswith(os.path.abspath('.')):
		return path.replace(os.path.abspath('.'), get_project_directory(), 1)
	return path

def get_webapp_taskrunner(webapp_root):
	if webapp_root.endswith('/'):
		parent = webapp_root.rsplit('/', 2)[0]
	else:
		parent = webapp_root.rsplit('/', 1)[0]
	task_runner = None
	if os.path.exists(parent + '/Gruntfile.js') or os.path.exists(parent + '/Gruntfile.coffee'):
		task_runner = 'grunt'
	elif os.path.exists(parent + '/gulpfile.js') or os.path.exists(parent + '/gulpfile.coffee'):
		task_runner = 'gulp'
	elif os.path.exists(parent + '/_config.yml'):
		task_runner = 'jekyll'
	return parent, task_runner

def build_webapp(webapp_root, task=None):
	"""Build a web app. Assumes that the webapp_root is a build subdirectory to a parent project with a build file."""
	parent, task_runner = get_webapp_taskrunner(webapp_root)
	if not task_runner:
		return
	def webapp_cmd(cmd):
		result = local(cmd)
		if result.failed:
			print red('Error running %s!' % cmd)
			raise UserWarning
	with cd_local(parent):
		# Install or update front-end packages
		if os.path.exists(os.path.join(parent, 'package.json')):
			if os.path.exists(os.path.join(parent, 'node_modules')):
				print bold('Updating npm packages ...')
				webapp_cmd('npm update')
			else:
				print bold('Installing npm packages ...')
				webapp_cmd('npm install')
		if os.path.exists(os.path.join(parent, 'bower.json')):
			if os.path.exists(os.path.join(parent, 'bower_components')):
				print bold('Updating bower packages ...')
				webapp_cmd('bower update')
			else:
				print bold('Installing bower packages ...')
				webapp_cmd('bower install')
		if os.path.exists(os.path.join(parent, 'config.js')):
			if os.path.exists(os.path.join(parent, 'jspm_packages')):
				print bold('Updating jspm packages ...')
				webapp_cmd('jspm update')
			else:
				print bold('Installing jspm packages ...')
				webapp_cmd('jspm install')
		# Run the build task
		print bold('Building %s ...' % parent)
		if not task:
			task = 'build'
		webapp_cmd('%s %s' % (task_runner, task))

_apt_get_install_set = set()

def apt_get_install(*packages):
	for package in packages:
		if package in _apt_get_install_set:
			continue
		_apt_get_install_set.add(package)
		# Check for the package
		with hide('everything'), settings(warn_only=True):
			result = run('dpkg -s %s' % package)
		if not result.succeeded:
			# TODO: separate the version from the package
			from hooks import hook
			with hook('apt-get install %s' % package):
				sudo('apt-get install %s -y' % package)
		else:
			print '%s already installed.' % package

def apt_get_update():
	# Don't update apt-get if not older than a day
	with hide('everything'), settings(warn_only=True):
		result = run('stat -c %Z /var/cache/apt/pkgcache.bin')
		if not env.get('setup') and result.succeeded and (int(time.time()) - int(result)) < (24 * 60 * 60):
			print 'apt-get is up to date'
			return
	sudo('apt-get update -y')

def apt_get_upgrade_packages():
	sudo('apt-get upgrade -y')

def get_project_directory():
	return os.path.join('/srv/www/apps', env['project'])

def get_manage_directory():
	return os.path.join('/srv/www/manage', env['project'])

def get_requirements_packages():
	try:
		with open('requirements.txt') as f:
			return tuple(map(lambda package: package.strip(), f.readlines()))
	except:
		return tuple()

def get_django_setting(site, setting):
	try:
		module = import_module(site['settings_module'])
		return getattr(module, setting)
	except:
		return None

def get_static_root(site):
	"""Get the STATIC_ROOT for a Django site. If not set and STATIC_ROOT is the default None then set to a production default."""
	static_root = get_django_setting(site, 'STATIC_ROOT')
	if not static_root:
		# Nevermind, static needs to be set in Django or it seems like it will throw an error
		return os.path.join('/srv/www/static', site['name'])
	return fix_absolute_path(static_root)

def get_media_root(site):
	"""Get the MEDIA_ROOT for a Django site. If not set and MEDIA_ROOT is the default '' then set to a production default."""
	media_root = get_django_setting(site, 'MEDIA_ROOT')
	if not media_root:
		return os.path.join('/srv/www/media', site['name'])
	return fix_absolute_path(media_root)

def get_webapp_root(site):
	"""Get the WEBAPP_ROOT but do not provide a production default if not set."""
	webapp_root = get_django_setting(site, 'WEBAPP_ROOT')
	if webapp_root:
		webapp_root = fix_absolute_path(webapp_root)
	return webapp_root

def get_static_url(site):
	"""Get the STATIC_URL for a Django site. If not set and STATIC_URL is the default None then set to a production default."""
	return get_django_setting(site, 'STATIC_URL') or '/static/'

def get_media_url(site):
	"""Get the MEDIA_URL for a Django site. If not set and MEDIA_URL is the default '' then set to a production default."""
	return get_django_setting(site, 'MEDIA_URL') or '/media/'

def get_webapp_url(site):
	return get_django_setting(site, 'WEBAPP_URL') or '/'

def get_virtual_env(site=None):
	# If virtualenv is available, then setup and use it
	with hide('everything'), settings(warn_only=True):
		if site and run('which virtualenv').succeeded:
			virtual_env = os.path.join('/srv/www/env', site['name'])
			if not exists(virtual_env):
				sudo('virtualenv ' + virtual_env)
			return virtual_env

def get_python_interpreter(site=None):
	virtual_env = get_virtual_env(site)
	if virtual_env:
		return os.path.join(virtual_env, 'bin/python')
	else:
		return 'python'

def get_pip_installer(site=None):
	virtual_env = get_virtual_env(site)
	if virtual_env:
		return os.path.join(virtual_env, 'bin/pip')
	else:
		with hide('everything'), settings(warn_only=True):
			if run('which pip', True).succeeded:
				return 'pip'
			else:
				return 'easy_install'

_pip_install_set = set()

def pip_install(site=None, *packages):
	# TODO: support upgrading packages with pip install --upgrade packagename
	# Check for pip first
	with hide('everything'), settings(warn_only=True):
		result = run('dpkg -s python-pip')
	if not result.succeeded:
		apt_get_install('python-pip')
	if site and len(packages) == 0:
		packages = tuple(site.get('pip_packages', [])) + get_requirements_packages()
	pip = get_pip_installer(site)
	# Install each listed package
	for package in packages:
		if site is None:
			if package in _pip_install_set:
				continue
			_pip_install_set.add(package)
		if package.startswith('git:') or package.startswith('git+'):
			apt_get_install('git')
		with hide('everything'), settings(warn_only=True):
			result = run('%s show %s' % (pip, package))
		# Check the output of pip show, it doesn't return non-zero on not finding the package, just no output
		if not len(result.strip()):
			# TODO: separate the version from the package for the hook
			from hooks import hook
			with hook('pip install %s' % package):
				# sudo with -H for setting the home directory for root so pip has proper permissions to cache
				run('sudo -H %s install "%s"' % (pip, package))
		else:
			print '%s already installed.' % package

def pip_update():
	# WARNING: Upgrading pip to the latest version will cause urllib3 warnings and errors to start occuring on Ubuntu 14
	# virtualenv uses the latest versions and will cause these warnings, but they are consumed in CompactStdout
	sudo('pip install --upgrade pip')
	sudo('pip install --upgrade distribute')

def set_environment(e):
	# Copy the environment name into each environments, even though only one is being used
	for name in environments:
		environments[name]['name'] = name
	# Apply the environment
	env.update(e)
	# Ensure that the sites dict exists
	if not env.get('sites'):
		env['sites'] = {}
	# Apply any default settings
	if environments.has_key('defaults'):
		for setting in environments['defaults']:
			if not env.has_key(setting):
				env[setting] = environments['defaults'][setting]
	# Apply default site settings to each site
	if env.has_key('sites') and env['sites'].has_key('defaults'):
		for site in env['sites'].values():
			for setting in env['sites']['defaults']:
				if not site.has_key(setting):
					site[setting] = env['sites']['defaults'][setting]
		del env['sites']['defaults']
	# Copy the site name into each of the sites and set the default type
	for name in env['sites']:
		env['sites'][name]['name'] = name
		if not env['sites'][name].has_key('type'):
			env['sites'][name]['type'] = SiteType.DJANGO

def set_default_environment(e):
	for arg in sys.argv:
		if arg.startswith('e:'):
			return
	set_environment(environments[e])
	env['default'] = True

def find_service(name):
	for service in env['services']:
		if service.name == name:
			return service
	return None



import tempfile, os

from fabric.api import env, sudo, settings
from fabric.contrib.files import upload_template

from .service import Service
from ..hooks import hook
from ..shared import apt_get_install, get_template, get_project_directory, get_virtual_env

class UWSGI(Service):
	name = 'uwsgi'
	script = 'uwsgi'

	def install(self):
		with hook('install %s' % self.name, self):
			apt_get_install('uwsgi', 'uwsgi-plugin-python')

	def get_custom_settings(self):
		lines = []
		for key, value in self.settings.items():
			if isinstance(value, bool):
				value = str(value).lower()
			lines.append('%s = %s' % (key, value))
		return '\n'.join(lines)

	def site_config(self, site):
		with hook('site config %s' % self.name, self, site):
			wsgi_module = site['wsgi_module'] if site.has_key('wsgi_module') else site['settings_module'].split('.')[0] + '.wsgi'
			context = {
				'project': env['project'],
				'project_dir': get_project_directory(),
				'settings_module': site['settings_module'],
				'wsgi_module': wsgi_module,
				'virtual_env': get_virtual_env(site),
				'custom_settings': self.get_custom_settings()
			}
			upload_template(get_template('uwsgi-app.ini'), '/etc/uwsgi/apps-available/%s.ini' % site['name'], context=context, use_sudo=True, mode=0644)
			sudo('chown root:root /etc/uwsgi/apps-available/%s.ini' % site['name'])
		self.restart()



from importlib import import_module
import os

from fabric.api import put, run, sudo, settings, env, local

from .hooks import hook
from .services.s3 import S3, get_aws_access_key
from .shared import *

def reboot():
	"""Reboot the server."""
	with hook('reboot'):
		sudo('shutdown -r now')

def install_key(key=None):
	"""Copy an ssh key to login later."""
	with settings(warn_only=True):
		run('mkdir -p .ssh')
		if not key:
			put('~/.ssh/id_rsa.pub', '.ssh/authorized_keys')
		else:
			put('~/.ssh/%s' % key, '.ssh/authorized_keys')

def generate_s3_ajax(module_name=''):
	"""Generate jquery ajax code for saving files with the S3 policy documents. A site must be specified."""
	# http://aws.amazon.com/articles/1434
	# http://docs.aws.amazon.com/AmazonS3/2006-03-01/dev/HTTPPOSTForms.html
	site = env.get('site')
	if not site:
		print red('Error: An environment and site must both first be specified.')
		exit(1)
	for s3_service in site['services']:
		if type(s3_service) is S3:
			break
		s3_service = None
	if not s3_service:
		print red('Error: There is no S3 service associated with this site.')
		exit(1)
	site_access_key_id, site_secret_access_key = get_aws_access_key(site)
	from django.conf import settings
	from django.template import Template, Context
	import json, hmac, hashlib, base64
	settings.configure()
	with open('%s/templates/s3-ajax.js' % os.path.dirname(__file__)) as f:
		template = Template(f.read())
	context = {'policies': [], 'module_name': module_name}
	policy_context_vars = ('success_action_redirect', 'bucket', 'acl')
	for name, policy in s3_service.settings['policy_documents'].items():
		aws_access_key_id = policy.pop('access_key_id', site_access_key_id)
		aws_secret_access_key = policy.pop('secret_access_key', site_secret_access_key)
		policy_str = json.dumps(policy)
		policy['access_key_id'] = aws_access_key_id
		policy['headers'] = {}
		policy['meta_data'] = []
		# Process the conditions
		for condition in policy['conditions']:
			if type(condition) is dict:
				key = condition.keys()[0]
				value = condition.values()[0]
			elif type(condition) in (list, tuple):
				# The first value will always be starts-with or eq but they are treated the same
				if condition[0] not in ('starts-with', 'eq'):
					continue
				key = condition[1][1:]
				value = condition[2]
				if key == 'key':
					policy['directory'] = value
					continue
			if key in policy_context_vars:
				policy[key] = value
			elif key == 'Expires' or key.startswith('Content-'):
				policy['headers'][key] = value
			elif key.startswith('x-amz-meta-'):
				policy['meta_data'].append(key[11:])
		policy['name'] = name
		policy['encoded'] = base64.b64encode(policy_str)
		policy['signature'] = base64.b64encode(hmac.new(aws_secret_access_key, policy['encoded'], hashlib.sha1).digest())
		context['policies'].append(policy)
	print template.render(Context(context, autoescape=False))

_webapp_process = None

def _signal_handler(sig, frame):
	import signal, sys
	signal.signal(signal.SIGINT, signal.SIG_IGN)
	if _webapp_process:
		_webapp_process.kill()
	sys.exit(0)

def run_webapp():
	"""Run a webapp. Serve up a stand-alone webapp site on port 8000. Assumes that the webapp_root is a build subdirectory to a parent project with a build file."""
	site = env.get('site')
	if not site:
		print red('Error: An environment and site must both first be specified.')
		exit(1)
	if not site.has_key('webapp') or not site['webapp'].get('root'):
		print red('Error: Site must be a stand-alone webapp with a root specified.')
		exit(1)
	import BaseHTTPServer, SimpleHTTPServer
	parent, task_runner = get_webapp_taskrunner(site['webapp']['root'])
	if task_runner == 'jekyll':
		build_webapp(site['webapp']['root'], site['webapp'].get('build_task'))
	elif task_runner:
		import subprocess, signal
		signal.signal(signal.SIGINT, _signal_handler)
		global _webapp_process
		print bold('Running %s ...' % parent)
		cwd = os.getcwd()
		os.chdir(parent)
		task = site['webapp'].get('run_task', 'runwebapp')
		_webapp_process = subprocess.Popen([task_runner, task])
		os.chdir(cwd)
	class WebAppRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler, object):
		def do_GET(self):
			# Strip any query arguments
			self.path = self.path.split('?')[0]
			# Convert to an index.html file
			if self.path.endswith('/'):
				self.path = self.path + 'index.html'
			# Try to rewrite as one of the extra webapp files
			files = site['webapp'].get('files')
			if files:
				dirname, filename = os.path.split(self.path)
				dirname = dirname.strip('/')
				if type(files) is dict:
					for prefix in files:
						if prefix.strip('/') == dirname:
							files = files[prefix]
							break
				if isinstance(files, (list, tuple)):
					for f in files:
						if filename == os.path.basename(f):
							self.path = '/%s' % f
							return super(WebAppRequestHandler, self).do_GET()
			# Rewrite to the webapp root
			self.path = '/' + site['webapp']['root'].strip('/') + self.path
			return super(WebAppRequestHandler, self).do_GET()
	httpd = BaseHTTPServer.HTTPServer(('', 8000), WebAppRequestHandler)
	print "Serving on port %d" % 8000
	httpd.serve_forever()

def uname():
	"""Print the remote uname."""
	run('uname -a')

def ntp():
	"""Sync the time on the host using network time."""
	sudo('ntpdate pool.ntp.org')

def exports():
	"""Prints commands to export AWS environment variables. Example use: eval `fab --hide running,status e:production x:exports`"""
	site = env.get('site')
	if not site:
		print red('Error: An environment and site must both first be specified.')
		exit(1)
	else:
		aws_access_key_id, aws_secret_access_key = get_aws_access_key(site)
		# Command line interface variables
		print 'export AWS_ACCESS_KEY=%s' % aws_access_key_id
		print 'export AWS_SECRET_KEY=%s' % aws_secret_access_key
		# Boto variables
		print 'export AWS_ACCESS_KEY_ID=%s' % aws_access_key_id
		print 'export AWS_SECRET_ACCESS_KEY=%s' % aws_secret_access_key

# Archive functions

def __git_archive():
	"""Creates a git archive with all submodules included. Uses gnutar so the archive is compatible on linux."""
	project = env.get('project', 'project')
	local("git archive HEAD -o %s.tar; git submodule foreach 'git archive --prefix ${path}/ HEAD -o ../temp.tar; gnutar -Af ../%s.tar ../temp.tar; rm ../temp.tar'; gzip -f %s.tar" % (project, project, project))

def __git_raw_archive():
	"""Creates an archive of files not based on the git repository. For testing uncommitted code."""
	project = env.get('project', 'project')
	local("gnutar -cf %s.tar .; gzip -f %s.tar" % (project, project))

def archive(raw=False):
	"""Creates a git archive with all submodules included. Uses gnutar so the archive is compatible on linux. Optional argument: true for a raw archive."""
	if raw:
		__git_raw_archive()
	else:
		__git_archive()

_MYSQL_CLIENT_INSTRUCTIONS = """1. Download the connector from: http://dev.mysql.com/downloads/connector/c/
2. Install the connector into /usr/local/
3. Add the correct /usr/local/mysql-connector-c-<your version here>/bin directory to PATH
4. Add the correct /usr/local/mysql-connector-c-<your version here>/lib directory to DYLD_LIBRARY_PATH"""

_MYSQL_PYTHON_INSTRUCTIONS = """1. Download MySQL-python from: https://pypi.python.org/pypi/MySQL-python
2. export CFLAGS=-Wunused-command-line-argument-hard-error-in-future
3. python setup.py build
4. sudo python setup.py install
"""

_PYCRYPTO_INSTRUCTIONS = """1. Download pycrypto from: https://www.dlitz.net/software/pycrypto/
2. export CFLAGS=-Wunused-command-line-argument-hard-error-in-future
3. ./configure
4. python setup.py build
5. sudo python setup.py install
"""

def local_setup():
	"""Installs packages necessary for a local mac setup."""
	with settings(warn_only=True):
		if env.has_key('pip_packages'):
			packages = tuple(env['pip_packages']) + DJANGO_PIP_PACKAGES
		else:
			packages = DJANGO_PIP_PACKAGES
		# Check to make sure mysql client libraries are installed
		if 'mysql-python' in packages:
			result = local('which mysql_config', True)
			if not result.succeeded:
				# Print mysql client setup instructions and exit
				print red('You do not have the mysql client libraries installed.')
				print 'Please follow these instructions first:'
				print _MYSQL_CLIENT_INSTRUCTIONS
				return

			try:
				import MySQLdb
			except:
				import platform
				if platform.system() == 'Darwin':
					print red('You do not have MySQL-python installed.')
					print 'Please install first from easy_install, pip, or the instructions below:'
					print _MYSQL_PYTHON_INSTRUCTIONS
					return

		try:
			import Crypto
		except:
			import platform
			if platform.system() == 'Darwin':
				print red('You do not have pycrypto installed.')
				print 'Please install first from easy_install, pip, or the instructions below:'
				print _PYCRYPTO_INSTRUCTIONS
				return

		print 'The following packages will be installed:'
		print '\n'.join(packages)
		print 'Continue? y/n:'
		i = None
		while not i: i = raw_input().lower()
		if i[0] != 'y':
			return
		result = local('which pip', True)
		if result.succeeded:
			for package in packages:
				local('sudo pip install %s' % package)
				if package == 'distribute':
					local('sudo pip install --upgrade distribute')
		else:
			for package in packages:
				if package.startswith('git+git'):
					package = package.replace('git+git', 'https', 1)
					package = package.replace('.git#', '/tarball/master#', 1)
				local('sudo easy_install %s' % package)
				if package == 'distribute':
					local('sudo easy_install -U distribute')
		print ''
		print bold('Note: if you have webapps with this project, these need to be setup seperately - potentially requiring nodejs and running npm install.')



#!%(interpreter)s
import os
import sys

if __name__ == '__main__':
	# NOTE: os.chdir() to the project does not work so instead add to sys.path
	sys.path.append('%(project_dir)s')
	os.environ.setdefault('DJANGO_SETTINGS_MODULE', '%(settings_module)s')
	from django.core.management import execute_from_command_line
	execute_from_command_line(sys.argv)



from django.contrib.humanize.templatetags.humanize import intcomma
from django import template
from django.template.defaultfilters import floatformat

register = template.Library()

@register.filter(is_safe=True)
def floatcomma(value, arg=-1):
	"""Combines the intcomma and floatformat filters into one."""
	i = intcomma(value)
	f = floatformat(value, arg)
	index = i.find('.')
	if index != -1:
		i = i[:index]
	index = f.find('.')
	if index != -1:
		f = f[index+1:].rstrip('0')
		if f:
			return '%s.%s' % (i, f)
	return i



import os
import subprocess
import sys

from fabric.api import env, sudo, run, hide, settings
from fabric.contrib.files import exists

# Dictionary of environments set from the fab file.
environments = {}

class SiteType:
	DJANGO = 1
	WEBAPP = 2
	NODE = 3
	SCGI = 4
	NGINX = 5

presets = {
	'GZIP_CONTENT_TYPES': ('text/html', 'text/css','application/javascript','application/x-javascript', 'image/svg+xml'),
	'NO_CACHE': 'no-cache',
	'PERMANENT_CACHE': 'max-age=315360000'
}

DJANGO_PACKAGES = ('python-pip', 'python-dev', 'sqlite3'),
DJANGO_PIP_PACKAGES = ('django', 'django-storages', 'boto', 'distribute')

# fabric.colors doesn't have bold, so just define styles here
def bold(msg):
	return '\033[1m%s\033[0m' % msg

def red(msg):
	return '\033[1;31m%s\033[0m' % msg

def green(msg):
	return '\033[1;32m%s\033[0m' % msg

def blue(msg):
	return '\033[1;34m%s\033[0m' % msg

def teal(msg):
	return '\033[1;36m%s\033[0m' % msg

def get_template_dir():
	return '%s/templates' % os.path.dirname(__file__)

def fix_static_path(path, site):
	"""Changes a relative path (os.path.abspath('.')) that may have been specified in the settings."""
	if path.startswith(os.path.abspath('.')):
		if site.get('static_inside_project', False):
			return path.replace(os.path.abspath('.'), '/srv/www/%s' % env['project'], 1)
		else:
			return path.replace(os.path.abspath('.'), '/srv/www', 1)
	return path

def fix_webapp_path(path):
	"""Like fix_static_path, but webapps should always stay inside the project."""
	if path.startswith(os.path.abspath('.')):
		return path.replace(os.path.abspath('.'), '/srv/www/%s' % env['project'], 1)
	return path

def get_webapp_taskrunner(webapp_root):
	if webapp_root.endswith('/'):
		parent = webapp_root.rsplit('/', 2)[0]
	else:
		parent = webapp_root.rsplit('/', 1)[0]
	task_runner = None
	if os.path.exists(parent + '/Gruntfile.js') or os.path.exists(parent + '/Gruntfile.coffee'):
		task_runner = 'grunt'
	elif os.path.exists(parent + '/gulpfile.js') or os.path.exists(parent + '/gulpfile.coffee'):
		task_runner = 'gulp'
	elif os.path.exists(parent + '/_config.yml'):
		task_runner = 'jekyll'
	return parent, task_runner

def build_webapp(webapp_root, task=None):
	"""Build a web app. Assumes that the webapp_root is a build subdirectory to a parent project with a build file."""
	parent, task_runner = get_webapp_taskrunner(webapp_root)
	if task_runner:
		print bold('Building %s ...' % parent)
		cwd = os.getcwd()
		os.chdir(parent)
		if not task:
			task = 'build'
		result = subprocess.call([task_runner, task])
		os.chdir(cwd)
		if result != 0:
			print red('Error: build failed!')
			raise UserWarning

_apt_get_install_set = set()

def apt_get_install(*packages):
	for package in packages:
		if package in _apt_get_install_set:
			continue
		_apt_get_install_set.add(package)
		# Check for the package
		with hide('everything'), settings(warn_only=True):
			result = run('dpkg -s %s' % package)
		if not result.succeeded:
			# TODO: separate the version from the package
			from hooks import hook
			with hook('apt-get install %s' % package):
				sudo('apt-get install %s -y' % package)
		else:
			print '%s already installed.' % package

def apt_get_update():
	sudo('apt-get update -y')

_pip_install_set = set()

def pip_install(*packages):
	for package in packages:
		if package in _pip_install_set:
			continue
		_pip_install_set.add(package)
		if package.startswith('git:') or package.startswith('git+'):
			apt_get_install('git')
		with hide('everything'), settings(warn_only=True):
			result = run('pip show %s' % package)
		# Check the output of pip show, it doesn't return non-zero on not finding the package, just no output
		if not len(result.strip()):
			# TODO: separate the version from the package
			from hooks import hook
			with hook('pip install %s' % package):
				sudo('pip install %s' % package)
		else:
			print '%s already installed.' % package

def pip_update():
	sudo('pip install --upgrade pip')
	sudo('pip install --upgrade distribute')

def set_environment(e):
	# Copy the environment name into each environments, even though only one is being used
	for name in environments:
		environments[name]['name'] = name
	# Apply the environment
	env.update(e)
	# Ensure that the sites dict exists
	if not env.get('sites'):
		env['sites'] = {}
	# Apply any default settings
	if environments.has_key('defaults'):
		for setting in environments['defaults']:
			if not env.has_key(setting):
				env[setting] = environments['defaults'][setting]
	# Apply default site settings to each site
	if env.has_key('sites') and env['sites'].has_key('defaults'):
		for site in env['sites'].values():
			for setting in env['sites']['defaults']:
				if not site.has_key(setting):
					site[setting] = env['sites']['defaults'][setting]
		del env['sites']['defaults']
	# Copy the site name into each of the sites and set the default type
	for name in env['sites']:
		env.sites[name]['name'] = name
		if not env.sites[name].has_key('type'):
			env.sites[name]['type'] = SiteType.DJANGO

def set_default_environment(e):
	for arg in sys.argv:
		if arg.startswith('e:'):
			return
	set_environment(environments[e])

def find_service(name):
	for service in env['services']:
		if service.name == name:
			return service
	return None



from importlib import import_module
import os
import signal
import subprocess

from django.conf import settings
from django.contrib.staticfiles.management.commands.runserver import Command as RunserverCommand
from utils.colors import bold, red
from utils.urls import static_serve_url

__webapp_process = None

def _signal_handler(sig, frame):
	"""Kill the runwebapp process on Control-C. Probably redundant because child processes should be automatically killed by the OS."""
	signal.signal(signal.SIGINT, signal.SIG_IGN)
	if __webapp_process:
		__webapp_process.kill()
	import sys
	sys.exit(0)

def get_webapp_taskrunner(webapp_root):
	"""Return gulp or gruntfile as the taskrunner if applicable."""
	if webapp_root.endswith('/'):
		parent = webapp_root.rsplit('/', 2)[0]
	else:
		parent = webapp_root.rsplit('/', 1)[0]
	task_runner = None
	if os.path.exists(parent + '/Gruntfile.js') or os.path.exists(parent + '/Gruntfile.coffee'):
		task_runner = 'grunt'
	elif os.path.exists(parent + '/gulpfile.js') or os.path.exists(parent + '/gulpfile.coffee'):
		task_runner = 'gulp'
	return parent, task_runner

def run_webapp(webapp_root):
	"""Run a webapp, based on a WEBAPP_ROOT setting.

	Assumes that the webapp_root is a build subdirectory to a parent project with a grunt file.
	"""
	global __webapp_process
	parent, task_runner = get_webapp_taskrunner(webapp_root)
	if task_runner:
		print bold('Running %s ...' % parent)
		cwd = os.getcwd()
		os.chdir(parent)
		__webapp_process = subprocess.Popen([task_runner, 'runwebapp'])
		os.chdir(cwd)

runwebapp_help = """
Starts a development web server for running a project, static files, and a single static web app (just like adding an additional STATIC_ROOT and STATIC_URL only for a specific web app with the settings below).
If a Gruntfile or gulpfile is in the parent directory, the webapp will first call the runwebapp task in the background, this should build, watch, and setup a livereload server.
The only possible non-development webapp setup is hosting on S3, with DNS pointing to the S3 bucket, and S3 redirecting 404, 403 to the django host if needed to share the same domain.
Currently, fab deployment will not automatically setup or configure the S3 bucket for the webapp

Requests for the non-development webapp must first go through the S3 bucket because:
  - Redirecting (rewrite) from nginx to S3 cannot be done because '/' is a typical webapp url, but '/' is needed for django
  - Serving statically (alias) from nginx cannot be done because webapps will typically reference other static files not under their directory, which this command and fab deploy resolves
  - Statically serving files with django, like this command is bad for non-development setups

Settings:
  WEBAPP_ROOT
                        e.g. os.path.abspath('.') + '/projectswebapp/build/'
                        Only used with this command and fab deploy
                        Will be converted to the correct path, like MEDIA_ROOT, when deploying with fab

  WEBAPP_URL
                        '/' (default) or '/projectswebapp/' used with this command only

  WEBAPP_INDEX
                        'index.html' (default) or some other filename
                        The index file to use when a directory is requested
                        For use with this command only
"""

class Command(RunserverCommand):
	help = runwebapp_help

	def handle(self, *args, **options):
		"""Stardard handle method. This runs on both the parent autoreload process and the main child process."""

		# Only run the taskrunner once from the parent autoreload process (where RUN_MAIN is None) that watches files and launches/relaunches the main webserver child process
		# https://github.com/django/django/blob/master/django/utils/autoreload.py
		if not os.environ.get("RUN_MAIN"):
			if not hasattr(settings, 'WEBAPP_ROOT') or not hasattr(settings, 'WEBAPP_URL'):
				print red('Error: There is no web app setup for this project!', True)
				raise UserWarning

			# Run webapp if applicable
			signal.signal(signal.SIGINT, _signal_handler)
			run_webapp(settings.WEBAPP_ROOT)

		return super(Command, self).handle(*args, **options)

	def get_handler(self, *args, **options):
		"""Return the RunserverCommand get_handler. This doesn't run on the parent autoreload process."""

		# Modify the urls
		webapp_url = getattr(settings, 'WEBAPP_URL', '/')
		module = import_module(settings.ROOT_URLCONF)
		module.urlpatterns.append(static_serve_url(webapp_url, settings.WEBAPP_ROOT, view='utils.views.webapp_serve'))

		return super(Command, self).get_handler(*args, **options)



from fabric.api import *
from fabric.contrib.files import sed, exists, upload_template, append
from colors import red, bold
from importlib import import_module
import tempfile, sys, os, urlparse, fnmatch

""" Environment keys:
install_key - True/False/string, copy and install the default id_rsa.pub ssh key or specified key
packages - override of the standard packages to install on setup
pip_packages - override of the standard pip packages to install on setup
nginx_settings - dict of settings to override global settings in /etc/nginx/nginx.conf, to be placed in /etc/nginx/conf.d/fabric.conf, e.g. {'http': {...}}
geoip - True/False, install and setup geoip for django
mysql_server - True/False, install and setup server as a mysql server
mysql_settings - dict of settings to override global settings in /etc/mysql/my.cnf, to be placed in /etc/mysql/conf.d/fabric.cnf, e.g. { 'mysqld': {'bind-address': '0.0.0.0'}}
postgresql_client - True/False, install and setup django as a postgresql client
postgresql_server - True/False, install and setup server as a postgresql server
postgresql_settings - dict of settings to add/override for the postgresql server, to be placed in /etc/postgresql/9.1/main/fabric.conf
geo_django - True/False, install and setup postgresql client/servers with geo capabilities
memcached_server - True/False/dict of settings to edit into /etc/memcached.conf, e.g. {'-M': None, '-c': '1024'}
redis_server - True/False/dict of settings to edit into /etc/redis.conf, e.g. {}
database_only - True/False, this server is a database server only (mysql, memcached, etc..) no django
project - name of the project, for use in the filesystem, should only be one project even with multiple sites
sites - array of dict settings for running multiple sites, with the following keys, these settings may also optionally can be specified on the root level if only one site
{
	site - domain of the server, used to setup nginx if ALLOWED_HOSTS is not set in the settings and is used as a key when choosing a site
	ssl_cert - ssl certificate to use
	ssl_cert_key - private key to go with the cert, both must be given
	ssl_only - if True, will configure nginx to redirect all port 80 traffic to https
	local_tests - list of django tests to run locally before a deploy
	remote_tests - list of django tests to run remotely after a deploy
	settings_module - the settings module to use on deploy etc.. (DJANGO_SETTINGS_MODULE), required
	crontab - list of jobs to add to the crontab for this site, each job is (m, h, dom, moy, dow, command)
	default - if True, this will be the default nginx site
	nginx_settings - dictionary of settings to set in the / location for this site
	custom_locations - array of custom location directives to add to the nginx site configuration
	static_inside_project - if True, static and media roots will remain inside the project when fixing their paths for the server. Otherwise media and static will go into /srv/www/media and /srv/www/static
	*project - automatically copied from env['project']
	webapp_bucket - the bucket to deploy to, does not configure the bucket itself.
	webapp_files - array of static files that the webapp uses that are not under the webapp directory, specified as relative static paths that can be passed to the findstatic command e.g. js/script.js
		A better way to get extra files is to symbolically link the shared directories into the webapp project, and if using gulp only match the required files when building
		Try to NOT use this.
	webapp_clear - clear the destination path in the bucket before deploying
	webapp_gzip_types - gzip webapp files of these types. Don't specify images and PDFs because they are already compressed (see GZIP_CONTENT_TYPES) (https://developer.yahoo.com/performance/rules.html)
	webapp_cache_control - Set the Cache-Control for the file patterns specified (uses fnmatch). For no caching, use NO_CACHE, for permanent caching use PERMANENT_CACHE. For versioned files, e.g. bundle.103GA74.css use permanent caching.
}
"""

""" Requirements:
These tasks expect media to be located at project/media and static to be at project/static 
"""

""" Notes:
memcached wouldn't restart properly without setting pty=False, so all calls to service have pty=False
"""

""" TODO:
mysql queries to run on setup, such as those required for opening up to allow clients.
http://www.thegeekstuff.com/2010/08/allow-mysql-client-connection/
"""

# Dictionary of environments set from the fab file.
environments = {}
# Like env, a site that should be used when running a few of the commands.
site = None

nginx_user = 'www-data'
postgresql_user = 'postgres'

# Standard packages, (includes sqlite, mysql, memcached client setup as standard along with boto for AWS)
basic_packages = ('gcc', 'make', 'linux-headers-$(uname -r)', 'build-essential', 'libtool', 'git', 'autoconf', 'zip')
standard_packages = basic_packages + ('nginx-full', 'uwsgi', 'uwsgi-plugin-python', 'python-pip', 'python-dev', 'sqlite3', 'mysql-client', 'libmysqlclient-dev')
pil_packages = ('libjpeg-dev', 'libpng-dev', 'libfreetype6-dev')
django_pip_packages = ('django', 'south', 'django-storages', 'boto', 'distribute', 'mysql-python', 'python-memcached')
pil_pip_packages = ('pil',)
celery_pip_packages = ('celery', 'kombu', 'django-celery')
facebook_pip_packages = ('facebook-sdk',)
facebook_django_pip_packages = ('git+git://github.com/pythonforfacebook/django-facebook.git#egg=django-facebook',)

# Default settings
memcached_settings = {'-d': None, 'logfile': '/var/log/memcached.log', '-m': '64', '-p': '11211', '-u': 'memcache', '-l': '127.0.0.1'}
redis_settings = {}

GZIP_CONTENT_TYPES = ('text/css','application/javascript','application/x-javascript')
NO_CACHE = 'no-cache'
PERMANENT_CACHE = 'max-age=315360000'

# Helper functions

def __get_domain(url):
	return urlparse.urlparse(url).netloc

def __get_path(url):
	return urlparse.urlparse(url).path

def __fix_static_path(path, site):
	""" Changes a relative path (os.path.abspath('.')) that may have been specified in the settings. """
	if path.startswith(os.path.abspath('.')):
		if site.get('static_inside_project', False):
			return path.replace(os.path.abspath('.'), '/srv/www/%s' % env['project'], 1)
		else:
			return path.replace(os.path.abspath('.'), '/srv/www', 1)
	return path

def __fix_webapp_path(path):
	""" Like __fix_static_path, but webapps should always stay inside the project. """
	if path.startswith(os.path.abspath('.')):
		return path.replace(os.path.abspath('.'), '/srv/www/%s' % env['project'], 1)
	return path

def __slash_wrap(path):
	if path.startswith('/'):
		if path.endswith('/'):
			return path
		else:
			return path + '/'
	else:
		if path.endswith('/'):
			return '/' + path
		else:
			return '/%s/' % path

def __setup_aws_access_key(site):
	try:
		module = import_module(site['settings_module'])
		# Boto variables
		os.environ.setdefault("AWS_ACCESS_KEY_ID", module.AWS_ACCESS_KEY_ID)
		os.environ.setdefault("AWS_SECRET_ACCESS_KEY", module.AWS_SECRET_ACCESS_KEY)
	except:
		print red('Error: Could not access AWS_ACCESS_KEY_ID or AWS_SECRET_ACCESS_KEY from settings.', True)
		exit(1)

def __upload_to_s3(site, bucket, directory=None, files=None, prefix=None):
	""" Uploads files to an s3 bucket. Upload either an entire directory with files=None, or specific files with and optional directory prefix. """
	if bucket is None:
		print red('Error: Bucket must be specified.', True)
		return
	if directory is None and files is None:
		print red('Error: Directory and/or files must be specified.', True)
		return
	# Setup boto
	import boto
	from boto.s3.bucket import Bucket
	from boto.s3.key import Key
	import mimetypes

	__setup_aws_access_key(site)

	# Connect to S3
	c = boto.connect_s3()
	b = Bucket(c, bucket)

	# Fix the prefix
	# prefix itself shouldn't have a / prefix itself but should end with /
	if prefix:
		prefix = prefix.lstrip('/')
		if prefix and not prefix.endswith('/'):
			prefix = prefix + '/'

	def __upload(key, filename):
		k = Key(b)
		k.key = key
		headers = {}
		content_type = mimetypes.guess_type(filename)[0]
		if site.has_key('webapp_cache_control'):
			for pattern in site['webapp_cache_control']:
				if fnmatch.fnmatch(filename, pattern):
					headers['Cache-Control'] = site['webapp_cache_control'][pattern]
					break
		if site.has_key('webapp_gzip_types') and content_type in site['webapp_gzip_types']:
			from gzip import GzipFile
			from StringIO import StringIO
			# Need to specify content_type when uploading from a string!
			headers['Content-Type'] = content_type
			headers['Content-Encoding'] = 'gzip'
			s = StringIO()
			g = GzipFile(fileobj=s, mode='wb')
			with open(filename, 'rb') as f:
				g.write(f.read())
			g.close()
			k.set_contents_from_string(s.getvalue(), headers)
		else:
			k.set_contents_from_filename(filename, headers)

	if files:
		# Upload individual files
		if directory:
			keys = [filename.lstrip('/') for filename in files]
			files = [os.path.join(directory, filename) for filename in files]
		else:
			keys = [os.path.split(filename)[1] for filename in files]
		for i, filename in enumerate(files):
			print 'Uploading ', keys[i]
			if prefix:
				key = prefix + keys[i]
			else:
				key = keys[i]
			__upload(key, filename)
	elif directory:
		# Upload an entire directory
		def __upload_dir(arg, dirname, names):
			# arg is the starting directory
			for name in names:
				filename = os.path.join(dirname, name)
				if not os.path.isdir(filename) and not os.path.islink(filename) and not name.startswith('.'):
					key = filename[len(arg):]
					if key.startswith('/'):
						key = key[1:]
					if prefix:
						key = prefix + key
					print 'Uploading ', key
					__upload(key, filename)
		os.path.walk(directory, __upload_dir, directory)

def __delete_from_s3(site, bucket, prefix=None):
	""" Remove all files with the prefix specified from the bucket. """
	if bucket is None:
		print red('Error: Bucket must be specified.', True)
		return
	# Setup boto
	import boto
	from boto.s3.bucket import Bucket
	from boto.s3.key import Key

	__setup_aws_access_key(site)

	# Fix the prefix
	if prefix:
		prefix = prefix.lstrip('/')

	# Connect to S3, list the contents, and remove all of the keys
	c = boto.connect_s3()
	b = Bucket(c, bucket)
	result_set = b.list(prefix=prefix)
	result = b.delete_keys([key.name for key in result_set])

# Package functions

def __apt_get_install(packages=standard_packages, update=True):
	if update:
		sudo('apt-get update -y')
	for package in packages:
		sudo('apt-get install %s -y' % package)

def __pip_install(packages=django_pip_packages, upgrade=True):
	if upgrade:
		sudo('pip install --upgrade pip')
		sudo('pip install --upgrade distribute')
	if 'pil' in packages:
		# http://stackoverflow.com/questions/7648200/pip-install-pil-e-tickets-1-no-jpeg-png-support
		with settings(warn_only=True):
			if exists('/usr/lib/x86_64-linux-gnu/libfreetype.so'):
				sudo('ln -s /usr/lib/x86_64-linux-gnu/libfreetype.so /usr/lib/libfreetype.so')
			if exists('/usr/lib/x86_64-linux-gnu/libz.so'):
				sudo('ln -s /usr/lib/x86_64-linux-gnu/libz.so /usr/lib/libz.so')
			if exists('/usr/lib/x86_64-linux-gnu/libjpeg.so'):
				sudo('ln -s /usr/lib/x86_64-linux-gnu/libjpeg.so /usr/lib/libjpeg.so')
	for package in packages:
		sudo('pip install %s' % package)

def __install_geoip():
	""" Install geoip (convert ip addresses into locations). """
	if not exists('/usr/local/geoip'):
		__apt_get_install(('libgeoip1',), False)
		run('wget --no-clobber http://www.maxmind.com/download/geoip/database/GeoLiteCity.dat.gz')
		run('gunzip GeoLiteCity.dat.gz')
		sudo('mkdir /usr/local/geoip')
		sudo('mv GeoLiteCity.dat /usr/local/geoip/')

def __create_postgresql_database_template_old():
	# Create the database template
	run('wget --no-clobber https://docs.djangoproject.com/en/dev/_downloads/create_template_postgis-1.5.sh')
	run('chmod 755 create_template_postgis-1.5.sh')
	sudo('su postgres -c ./create_template_postgis-1.5.sh')

def __install_postgis(old=False):
	__apt_get_install(('postgresql-server-dev-all', 'libpq-dev', 'libxml2', 'libxml2-dev'), False)

	# Install geos
	run('wget --no-clobber http://archive.ubuntu.com/ubuntu/pool/universe/g/geos/geos_3.3.3.orig.tar.gz')
	run('tar -xzf geos_3.3.3.orig.tar.gz')
	with cd('geos-3.3.3'):
		run('./configure')
		run('make')
		sudo('make install')

	# Install proj.4
	# https://docs.djangoproject.com/en/dev/ref/contrib/gis/install/
	run('wget --no-clobber http://download.osgeo.org/proj/proj-4.8.0.tar.gz')
	run('wget --no-clobber http://download.osgeo.org/proj/proj-datumgrid-1.5.tar.gz')
	run('tar -xzf proj-4.8.0.tar.gz')
	with cd('proj-4.8.0/nad'):
		run('tar -xzf ../../proj-datumgrid-1.5.tar.gz')
		with cd('..'):
			run('./configure')
			run('make')
			sudo('make install')

	# Install gdal
	run('wget --no-clobber http://download.osgeo.org/gdal/gdal-1.9.2.tar.gz')
	run('tar -xzf gdal-1.9.2.tar.gz')
	with cd('gdal-1.9.2'):
		run('./configure')
		run('make')
		sudo('make install')
		sudo('ldconfig')

	if old:
		run('wget --no-clobber http://postgis.refractions.net/download/postgis-1.5.5.tar.gz')
		run('tar -xzf postgis-1.5.5.tar.gz')
		with cd('postgis-1.5.5'):
			run('./configure')
			run('make')
			sudo('make install')
	else:
		run('wget --no-clobber http://download.osgeo.org/postgis/source/postgis-2.0.3.tar.gz')
		run('tar -xzf postgis-2.0.3.tar.gz')
		with cd('postgis-2.0.3'):
			run('./configure')
			run('make')
			sudo('make install')

def __setup_postgresql_database():
	# Create the user for django to access the database with
	for site in env['sites']:
		with settings(warn_only=True):
			module = import_module(site['settings_module'])
			DATABASES = module.DATABASES
			sudo('createuser --createdb --no-superuser --no-createrole %s' % DATABASES['default']['USER'], user=postgresql_user)
			sudo("psql -c \"ALTER USER %s WITH PASSWORD '%s';\"" % (DATABASES['default']['USER'], DATABASES['default']['PASSWORD']), user=postgresql_user)
			sudo('createuser --createdb --no-superuser --no-createrole %s' % env.user, user=postgresql_user)
			sudo('dropdb %s' % DATABASES['default']['NAME'], user=postgresql_user)
		sudo('createdb %s' % DATABASES['default']['NAME'], user=postgresql_user)
		if env.has_key('geo_django') and env['geo_django']:
			sudo("psql %s -c 'CREATE EXTENSION postgis;'" % DATABASES['default']['NAME'], user=postgresql_user)
			sudo("psql %s -c 'CREATE EXTENSION postgis_topology;'" % DATABASES['default']['NAME'], user=postgresql_user)
			# http://tmbu.blogspot.com/2012/11/postgres-postgis-django-os-x-and.html
			# http://stackoverflow.com/questions/4737982/django-avoids-creating-pointfield-in-the-database-when-i-run-python-manage-py-sy
			sudo("psql %s -c 'GRANT ALL PRIVILEGES ON geometry_columns TO %s;'" % (DATABASES['default']['NAME'], DATABASES['default']['USER']), user=postgresql_user)
			sudo("psql %s -c 'GRANT ALL PRIVILEGES ON spatial_ref_sys TO %s;'" % (DATABASES['default']['NAME'], DATABASES['default']['USER']), user=postgresql_user)
			# old
			#run('createdb -T template_postgis %s' % DATABASES['default']['NAME'])

def __setup_postgresql_settings():
	if env.has_key('postgresql_settings'):
		with tempfile.NamedTemporaryFile('w') as f:
			for setting in env['postgresql_settings']:
				f.write('%s = %s\n' % (setting, env['postgresql_settings'][setting]))
			f.flush()
			put(f.name, '/etc/postgresql/9.1/main/fabric.conf', use_sudo=True, mode=0644)
			sudo('chown %s:%s /etc/postgresql/9.1/main/fabric.conf' % (postgresql_user, postgresql_user))
			append('/etc/postgresql/9.1/main/postgresql.conf', "\n\ninclude 'fabric.conf'\n", use_sudo=True)
		sudo('service postgresql restart', pty=False)

def __setup_postgresql_server():
	""" Setup and install postgresql. """
	__apt_get_install(('postgresql',), False)
	if env.has_key('geo_django') and env['geo_django']:
		__install_postgis()
		#__create_postgresql_database_template_old()
	__setup_postgresql_database()
	__setup_postgresql_settings()

def __setup_postgresql_client():
	if env.has_key('geo_django') and env['geo_django'] and not (env.has_key('postgresql_server') and env['postgresql_server']):
		# Install PostGIS
		__install_postgis()
	# Install python postgresql support
	__apt_get_install(('python-dev', 'postgresql-server-dev-all'), False)
	__pip_install(('psycopg2',), False)

def __setup_mysql_settings():
	if env.has_key('mysql_settings'):
		with tempfile.NamedTemporaryFile('w') as f:
			for section in env['mysql_settings']:
				f.write('[%s]\n' % section)
				for setting in env['mysql_settings'][section]:
					f.write('%s = %s\n' % (setting, env['mysql_settings'][section][setting]))
			f.flush()
			put(f.name, '/etc/mysql/conf.d/fabric.cnf', use_sudo=True, mode=0644)		
			sudo('chown root:root /etc/mysql/conf.d/fabric.cnf')
		sudo('service mysql restart', pty=False)

def __setup_mysql():
	""" Install and setup mysql. """
	__apt_get_install(('mysql-server',), False)
	__setup_mysql_settings()
	# TODO: setup user, database etc... here

def __configure_memcached():
	if env.has_key('memcached_server') and type(env['memcached_server']) is dict:
		with tempfile.NamedTemporaryFile('w') as f:
			for setting in env['memcached_server']:
				if env['memcached_server'][setting]:
					f.write('%s %s\n' % (setting, env['memcached_server'][setting]))
				else:
					f.write('%s\n' % setting)
			f.flush()
			put(f.name, '/etc/memcached.conf', use_sudo=True, mode=0644)
			sudo('chown root:root /etc/memcached.conf')
		sudo('service memcached restart', pty=False)

def __setup_memcached():
	""" Install and setup memcached. """
	__apt_get_install(('memcached',), False)
	__configure_memcached()

def __configure_redis():
	if env.has_key('redis_server') and type(env['redis_server']) is dict:
		with tempfile.NamedTemporaryFile('w') as f:
			for setting in env['redis_server']:
				if env['redis_server'][setting]:
					f.write('%s %s\n' % (setting, env['redis_server'][setting]))
				else:
					f.write('%s\n' % setting)
			f.flush()
			put(f.name, '/etc/redis/fabric.conf', use_sudo=True, mode=0644)
			sudo('chown root:root /etc/redis/fabric.conf')
			append('/etc/redis/redis.conf', 'include fabric.conf', use_sudo=True)
		sudo('service redis-server restart', pty=False)

def __setup_redis():
	""" Install and setup memcached. """
	__apt_get_install(('redis-server',), False)
	__configure_redis()

NGINX_SSL = """listen 443 ssl;\n\tssl_certificate %s;\n\tssl_certificate_key %s;"""
NGINX_LOCATION = """location %s {\n\t\talias %s;\n\t\texpires 1d;\n\t}"""
NGINX_LOCATION_DOMAIN = """location %s {\n\t\trewrite ^(.*)$ http://%s$1 permanent;\n\t}"""
NGINX_WEBAPP_LOCATION = """location %s {\n\t\talias %s;\n\t\ttry_files $uri $uri/%s %s;\n\t}"""

def __setup_nginx():
	""" Run the nginx setup for all sites. """
	# Write new settings
	if env.has_key('nginx_settings') and type(env['nginx_settings']) is dict:
		with tempfile.NamedTemporaryFile('w') as f:
			for section in env['nginx_settings']:
				if type(env['nginx_settings'][section]) is dict:
					f.write('%s {\n' % section)
					for setting in env['nginx_settings'][section]:
						f.write('%s %s;\n' % (setting, env['nginx_settings'][section][setting]))
					f.write('}\n')
				elif type(env['nginx_settings'][section]) is bool:
					f.write('%s %s;\n' % (setting, 'on' if env['nginx_settings'][section] else 'off'))
				else:
					f.write('%s %s;\n' % (setting, str(env['nginx_settings'][section])))
			f.flush()
			put(f.name, '/etc/nginx/conf.d/fabric.conf', use_sudo=True, mode=0644)		
			sudo('chown root:root /etc/nginx/conf.d/fabric.conf')

	# Write the site settings
	for site in env['sites']:
		if site.has_key('default') and site['default']:
			site['default_str'] = ' default_server'
		else:
			site['default_str'] = ''
		if site.has_key('ssl_cert') and site.has_key('ssl_cert_key') and site['ssl_cert'] and site['ssl_cert_key']:
			site['ssl_str'] = NGINX_SSL % (site['ssl_cert'], site['ssl_cert_key'])
		else:
			site['ssl_str'] = ''
		module = import_module(site['settings_module'])
		if module.ALLOWED_HOSTS:
			site['allowed_hosts'] = ' '.join(module.ALLOWED_HOSTS)
		else:
			site['allowed_hosts'] = site['site']
		# Setup the static and media locations
		locations = []
		domain = __get_domain(module.STATIC_URL)
		if domain:
			locations.append(NGINX_LOCATION_DOMAIN % (__slash_wrap(__get_path(module.STATIC_URL)), domain))
		else:
			locations.append(NGINX_LOCATION % (__slash_wrap(module.STATIC_URL), __fix_static_path(module.STATIC_ROOT, site)))
		domain = __get_domain(module.MEDIA_URL)
		if domain:
			locations.append(NGINX_LOCATION_DOMAIN % (__slash_wrap(__get_path(module.MEDIA_URL)), domain))
		else:
			locations.append(NGINX_LOCATION % (__slash_wrap(module.MEDIA_URL), __fix_static_path(module.MEDIA_ROOT, site)))
		# Add any custom locations
		if site.has_key('custom_locations'):
			if isinstance(site['custom_locations'], (tuple, list)):
				for location in site['custom_locations']:
					locations.append(location)
			else:
				locations.append(site['custom_locations'])
		site['static_locations'] = '\n\n\t'.join(locations)
		if site.has_key('nginx_settings') and type(site['nginx_settings']) is dict:
			site['nginx_settings_str'] = '\n\t\t'.join(['%s %s;' % (setting, site['nginx_settings'][setting]) for setting in site['nginx_settings']])
		else:
			site['nginx_settings_str'] = ''
		# Configure the webapp if necessary
		webapp_root = getattr(module, 'WEBAPP_ROOT', None)
		webapp_url = __slash_wrap(getattr(module, 'WEBAPP_URL', '/'))
		if webapp_root:
			webapp_root = __fix_webapp_path(webapp_root)
			webapp_index = getattr(module, 'WEBAPP_INDEX', 'index.html')
			if webapp_url == '/':
				site['app_location'] = '@%s-app' % site['site'].replace('.', '-')
				site['webapp_location'] = NGINX_WEBAPP_LOCATION % (webapp_url, webapp_root, webapp_index, site['app_location'])
			else:
				site['app_location'] = '/'
				site['webapp_location'] = NGINX_WEBAPP_LOCATION % (webapp_url, webapp_root, webapp_index, '=404')
			# Configure the 404 file if available
			if os.path.isfile(os.path.join(module.WEBAPP_ROOT, '404.html')):
				site['webapp_404'] = 'error_page 404 %s404.html;' % webapp_url
			else:
				site['webapp_404'] = ''
		else:
			site['app_location'] = '/'
			site['webapp_location'] = ''
			site['webapp_404'] = ''
		if site.has_key('ssl_only') and site['ssl_only']:
			nginx_template = '%s/templates/nginx-site-https.conf' % os.path.dirname(__file__)
		else:
			nginx_template = '%s/templates/nginx-site.conf' % os.path.dirname(__file__)
		upload_template(nginx_template, '/etc/nginx/sites-available/%s.conf' % site['site'], context=site, use_sudo=True, mode=0644)
		sudo('chown root:root /etc/nginx/sites-available/%s.conf' % site['site'])
		with settings(warn_only=True):
			sudo('ln -s /etc/nginx/sites-available/%s.conf /etc/nginx/sites-enabled/%s.conf' % (site['site'], site['site']))
	sudo('service nginx restart', pty=False)

def __setup_uwsgi():
	""" Run the uwsgi setup for all sites. """
	for site in env['sites']:
		upload_template('%s/templates/uwsgi-app.ini' % os.path.dirname(__file__), '/etc/uwsgi/apps-available/%s.ini' % site['site'], context=site, use_sudo=True, mode=0644)
		with settings(warn_only=True):
			sudo('ln -s /etc/uwsgi/apps-available/%s.ini /etc/uwsgi/apps-enabled/%s.ini' % (site['site'], site['site']))
	sudo('service uwsgi restart', pty=False)

def create_public_bucket_policy(bucket_name):
	import json
	policy = {
		"Version": "2008-10-17",
		"Statement": [
			{
				"Sid": "AddPerm",
				"Effect": "Allow",
				"Principal": {
					"AWS": "*"
				},
				"Action": "s3:GetObject",
				"Resource": "arn:aws:s3:::%s/*" % bucket_name
			}
		]
	}
	return json.dumps(policy)

DEFAULT_CORS_RULE = {'allowed_method': ['GET'], 'allowed_origin': ['*'], 'allowed_header': ['Authorization'], 'max_age_seconds': 3000}

def __setup_buckets():
	""" Setup buckets on S3. If site is not specified, then the command will be run on all sites. """
	global site
	if not site:
		for site in env['sites']:
			if site.has_key('buckets'):
				__setup_buckets()
		site = None
		return

	__setup_aws_access_key(site)

	from boto import connect_s3
	from boto.s3.bucket import Bucket
	from boto.s3.key import Key

	for bucket_config in site['buckets']:
		# Connect and make sure the bucket exists
		print bold(u'Configuring bucket %s...' % bucket_config['name'])
		connection = connect_s3()
		try:
			bucket = connection.get_bucket(bucket_config['name'])
		except:
			connection.create_bucket(bucket_config['name'])
		# Set the bucket policy
		if bucket_config.has_key('policy'):
			bucket.set_policy(bucket_config['policy'])
		# Setup CORS, array of rules
		# http://boto.readthedocs.org/en/latest/ref/s3.html#boto.s3.cors.CORSConfiguration
		if bucket_config.has_key('cors') and bucket_config['cors'] is None:
			# If explicity set to None, then remove the cors policy
			bucket.delete_cors()
		else:
			if not bucket_config.has_key('cors'):
				# If not specified, use the default GET policy
				bucket_config['cors'] = (DEFAULT_CORS_RULE,)
			from boto.s3.cors import CORSConfiguration
			cors_config = CORSConfiguration()
			for rule in bucket_config['cors']:
				cors_config.add_rule(**rule)
			bucket.set_cors(cors_config)
		# Setup the lifecycle, array of rules
		# http://boto.readthedocs.org/en/latest/ref/s3.html#boto.s3.lifecycle.Lifecycle
		if bucket_config.has_key('lifecycle'):
			from boto.s3.lifecycle import Lifecycle
			lifecycle_config = Lifecycle()
			for rule in bucket_config['lifecycle']:
				lifecycle_config.add_rule(**rule)
			bucket.configure_lifecycle(lifecycle_config)
		else:
			bucket.delete_lifecycle_configuration()
		# Setup the bucket website hosting {suffix, error_key, routing_rules, redirect_all_requests_to}
		# http://boto.readthedocs.org/en/latest/ref/s3.html
		# https://github.com/boto/boto/blob/develop/boto/s3/website.py
		if bucket_config.has_key('website'):
			# Expand the routing rules, array of {condition, redirect}
			if bucket_config['website'].has_key('routing_rules'):
				from boto.s3.website import RoutingRules, RoutingRule
				routing_rules = RoutingRules()
				for rule in bucket_config['website']['routing_rules']:
					routing_rules.add_rule(RoutingRule(**rule))
				bucket_config['website']['routing_rules'] = routing_rules
			# Expand the redirect, redirect_all_requests_to is {hostname, protocol}
			if bucket_config['website'].has_key('redirect_all_requests_to'):
				from boto.s3.website import RedirectLocation
				bucket_config['website']['redirect_all_requests_to'] = RedirectLocation(**bucket_config['website']['redirect_all_requests_to'])
			bucket.configure_website(**bucket_config['website'])
		else:
			bucket.delete_website_configuration()

def __generate_s3_ajax():
	""" Generate jquery ajax code for saving files with the S3 policy documents. A site must be specified. """
	# http://aws.amazon.com/articles/1434
	# http://docs.aws.amazon.com/AmazonS3/2006-03-01/dev/HTTPPOSTForms.html
	global site
	if not site:
		print red('Error: An environment and site must both first be specified.')
		exit(1)
	module = import_module(site['settings_module'])
	from django.conf import settings
	from django.template import Template, Context
	import json, hmac, hashlib, base64
	settings.configure()
	with open('%s/templates/s3-ajax.js' % os.path.dirname(__file__)) as f:
		template = Template(f.read())
	context = {'access_key_id': module.AWS_ACCESS_KEY_ID, 'secret_access_key': module.AWS_SECRET_ACCESS_KEY, 'policies': []}
	policy_context_vars = ('success_action_redirect', 'bucket', 'acl')
	for name, policy in site['policy_documents'].items():
		policy_str = json.dumps(policy)
		policy['headers'] = {}
		# Process the conditions
		for condition in policy['conditions']:
			if type(condition) is dict:
				key = condition.keys()[0]
				value = condition.values()[0]
			elif type(condition) in (list, tuple):
				# The first value will always be starts-with or eq but they are treated the same
				if condition[0] not in ('starts-with', 'eq'):
					continue
				key = condition[1][1:]
				value = condition[2]
				if key == 'key':
					policy['directory'] = value
					continue
			if key in policy_context_vars:
				policy[key] = value
			elif key == 'Expires' or key.startswith('Content-'):
				policy['headers'][key] = value
		policy['name'] = name
		policy['encoded'] = base64.b64encode(policy_str)
		policy['signature'] = base64.b64encode(hmac.new(module.AWS_SECRET_ACCESS_KEY, policy_str, hashlib.sha1).digest())
		context['policies'].append(policy)
	print template.render(Context(context, autoescape=False))

def __set_environment(e):
	env.update(e)
	# Ensure that the sites array exists
	if not env.has_key('sites') or not env['sites']:
		env['sites'] = [dict(env)]
	# Copy the project name into each of the sites
	for site in env['sites']:
		site['project'] = env['project']

def set_default_environment(e):
	for arg in sys.argv:
		if arg.startswith('e:'):
			return
	__set_environment(environments[e])

# Common tasks

@task
def e(name):
	""" Set the environment for use with other commands. """
	if name and name.lower() in environments.keys():
		__set_environment(environments[name.lower()])
	else:
		print "Error: Unknown environment specified."

@task
def s(name):
	""" Set the site to use with other commands. Currently: manage and deploywebapp """
	global site
	if not env.has_key('sites'):
		print red('Error: An environment must first be specified.')
		exit(1)
	for s in env['sites']:
		if s['site'].lower().find(name.lower()) != -1:
			site = s
			break
		s = None

@task
def setup():
	""" Setup a server from scratch. """

	# Copy an ssh key to login later
	if env.has_key('install_key') and env['install_key']:
		with settings(warn_only=True):
			run('mkdir -p .ssh')
			if type(env['install_key']) is bool:
				put('~/.ssh/id_rsa.pub', '.ssh/authorized_keys')
			else:
				put('~/.ssh/%s' % env['install_key'], '.ssh/authorized_keys')

	if not env.has_key('database_only') or not env['database_only']:
		# Install packages
		if env.has_key('packages'):
			__apt_get_install(env['packages'])
		else:
			__apt_get_install()
		
		# Install pip packages
		if env.has_key('pip_packages'):
			__pip_install(env['pip_packages'])
		else:
			__pip_install()

		# Setup the web server
		__setup_nginx()
		__setup_uwsgi()

		# Install geoip library/data
		if env.has_key('geoip') and env['geoip']:
			__install_geoip()
	else:
		# Only a database server, make sure packages are up to date and install the basics
		sudo('apt-get update -y')
		if env.has_key('packages'):
			__apt_get_install(env['packages'])
		else:
			__apt_get_install(basic_packages)

	if env.has_key('mysql_server') and env['mysql_server']:
		__setup_mysql()

	if env.has_key('postgresql_server') and env['postgresql_server']:
		__setup_postgresql_server()

	if env.has_key('memcached_server') and env['memcached_server'] is not False:
		__setup_memcached()

	if env.has_key('postgresql_client') and env['postgresql_client']:
		__setup_postgresql_client()

@task
def upgrade():
	""" Upgrade a server's packages. """
	sudo('apt-get upgrade -y')
	if env.has_key('pip_packages'):
		packages = env['pip_packages']
	else:
		packages = django_pip_packages
	for package in packages:
		sudo('pip install --upgrade %s' % package)

@task
def restart():
	""" Restarts nginx and uWSGI. """
	sudo('service uwsgi restart')
	sudo('service nginx restart')

@task
def manage(*args):
	""" Run a management command using the supplied arguments. If site is not specified command will be run on all sites. """

	if site is not None:
		with cd('/srv/www/%s' % site['project']):
			sudo('python manage.py %s --settings %s' % (' '.join(args), site['settings_module']), user=nginx_user)
	else:
		for site in env['sites']:
			with cd('/srv/www/%s' % site['project']):
				sudo('python manage.py %s --settings %s' % (' '.join(args), site['settings_module']), user=nginx_user)

class own_project(object):
	""" Changes the permissions of the project between the ssh user and nginx user. """
	def __init__(self):
		pass
	def __enter__(self):
		sudo('chown -R %s:%s /srv/www/%s' % (env['user'], env['user'], env['project']))
	def __exit__(self, *_):
		sudo('chown -R %s:%s /srv/www/%s' % (nginx_user, nginx_user, env['project']))

def __get_excluded():
	"""
	Get an excluded list of patterns.
	rsync patterns are similar to .gitignore patterns. e.g. /foo matches foo at the base.
	http://git-scm.com/docs/gitignore
	http://ss64.com/bash/rsync.html
	"""
	excluded = ['.*', '*.pyc', '*.db']
	# Exclude anything in the root .gitignore file
	if os.path.exists('.gitignore'):
		with file('.gitignore', 'r') as f:
			for line in f:
				line = line.strip()
				if line.startswith('.') or line.find('!') != -1:
					continue
				if line not in excluded:
					excluded.append(line)
	# Exlude all of the webapp's node_modules
	for site in env['sites']:
		module = import_module(site['settings_module'])
		os.environ.setdefault("DJANGO_SETTINGS_MODULE", site['settings_module'])
		if hasattr(module, 'WEBAPP_ROOT'):
			try:
				from management.commands.runwebapp import get_webapp_taskrunner
				parent, task_runner = get_webapp_taskrunner(module.WEBAPP_ROOT)
				if task_runner:
					excluded.append('/' + os.path.relpath(parent) + '/node_modules')
					excluded.append('/' + os.path.relpath(parent) + '/bower_components')
				# If wanting to exclude the entire webapp
				#if task_runner:
				#	excluded.append('/' + os.path.relpath(parent))
				#else:
				#	excluded.append('/' + os.path.relpath(module.WEBAPP_ROOT))
			except:
				print red('Warning: Could not import webapp settings for %s when syncing.' % site['site'], True)
	return excluded

SYNC_COMMAND = 'rsync -avz %s --delete --delete-excluded%s ./ %s@%s:/srv/www/%s'

def sync():
	""" Syncs the local code to the server. Used as part of the deploy process. Wrap in the task decorator to enable. sync = task(sync) """
	sudo('mkdir -p /srv/www/%s' % env['project'])
	with own_project():
		# Preserve existing media if a subdirectory of the project
		for site in env['sites']:
			module = import_module(site['settings_module'])
			MEDIA_ROOT = __fix_static_path(module.MEDIA_ROOT, site)
			if MEDIA_ROOT.startswith('/srv/www/%s/' % env['project']) and exists(MEDIA_ROOT):
				sudo('mv %s /srv/www/%smedia' % (MEDIA_ROOT, env['project']))

		# Sync and set the owner
		excluded = __get_excluded()
		excluded = ['--exclude="%s"' % ex for ex in excluded]
		local(SYNC_COMMAND % (' '.join(excluded), ' -e "ssh -i %s"' % os.path.expanduser(env['key_filename']) if env.get('key_filename') else '', env['user'], env['hosts'][0], env['project']))

		# Create the media directory
		# Restore existing media if a subdirectory of the project, if it is outside the project make sure it is owned by nginx
		for site in env['sites']:
			module = import_module(site['settings_module'])
			MEDIA_ROOT = __fix_static_path(module.MEDIA_ROOT, site)
			sudo('mkdir -p %s' % MEDIA_ROOT)
			if MEDIA_ROOT.startswith('/srv/www/%s/' % env['project']):
				if exists('/srv/www/%smedia' % env['project']):
					# Delete the media directory but leave intermediate directories then restore with a move
					sudo('rm -rf %s' % MEDIA_ROOT)
					sudo('mv /srv/www/%smedia %s' % (env['project'], MEDIA_ROOT))
			else:
				sudo('chown -R %s:%s %s' % (nginx_user, nginx_user, MEDIA_ROOT))

		# Copy any additional webapp files
		if site.has_key('webapp_files') and site['webapp_files']:
			from django.contrib.staticfiles import finders
			module = import_module(site['settings_module'])
			if hasattr(module, 'WEBAPP_ROOT'):
				WEBAPP_ROOT = __fix_webapp_path(module.WEBAPP_ROOT)
				for filename in site['webapp_files']:
					result = finders.find(filename)
					if not result:
						print red('Error: Could not find static file %s.' % filename, True)
						return
					result = result[0] if isinstance(result, (list, tuple)) else result
					put(result, os.path.join(WEBAPP_ROOT, filename), use_sudo=True, mode=0644)

def __sync_dry_run():
	""" Do an rsync dry run to see which files will be updated when deploying. """
	# e.g. To show just migrations: fab e:production x:dryrun | grep -v "^deleting" | grep -v "/$" | grep "^shared/migrations"
	excluded = __get_excluded()
	excluded = ['--exclude="%s"' % ex for ex in excluded]
	local(SYNC_COMMAND % (' '.join(excluded), ' --dry-run -e "ssh -i %s"' % os.path.expanduser(env['key_filename']) if env.get('key_filename') else '', env['user'], env['hosts'][0], env['project']))

def sync_down(path=''):
	""" Syncs down stuff from the server. Wrap in the task decorator to enable. sync_down = task(sync_down) """
	if path.startswith('/'):
		path = path[1:]
	if not path.endswith('/'):
		path += '/'
	local('rsync -avz --exclude=".*" --exclude="*.pyc" --exclude="*.sh" --exclude="*.db" %s@%s:/srv/www/%s%s ./%s' % (env['user'], env['hosts'][0], env['project'], path, path))

def __deploy_webapp():
	""" Deploy a webapp to S3 with the prefix of WEBAPP_URL. If site is not specified, then the command will be run on all sites. """
	global site
	if not site:
		for site in env['sites']:
			if site.has_key('webapp_bucket'):
				__deploy_webapp()
		site = None
		return

	try:
		module = import_module(site['settings_module'])
		root = module.WEBAPP_ROOT
		prefix = getattr(module, 'WEBAPP_URL', '/')
	except:
		print red('Error: Could not import webapp settings to deploy.', True)
		return
	# Build the webapp if needed
	os.environ.setdefault("DJANGO_SETTINGS_MODULE", site['settings_module'])
	from management.commands.runwebapp import build_webapp
	build_webapp(root)
	# Clear the bucket
	if site.has_key('webapp_clear') and site['webapp_clear']:
		__delete_from_s3(site, site['webapp_bucket'], prefix=prefix)
	# Upload the webapp files
	__upload_to_s3(site, site['webapp_bucket'], root, prefix=prefix)
	# Upload the static files found in other directories wih the static file finder
	if site.has_key('webapp_files') and site['webapp_files']:
		from django.contrib.staticfiles import finders
		for filename in site['webapp_files']:
			result = finders.find(filename)
			if not result:
				print red('Error: Could not find static file %s.' % filename, True)
				return
			result = result[0] if isinstance(result, (list, tuple)) else result
			__upload_to_s3(site, site['webapp_bucket'], result[:-len(filename)], (filename,), prefix=prefix)

def deploy_crontab():
	""" Creates a new crontab for the nginx user. """
	# list of (5,6,'*',0,'full shell command') or just 'management command'
	# http://www.adminschoice.com/crontab-quick-reference/

	# Get the existing crontab
	result = sudo('crontab -l -u %s' % nginx_user, warn_only=True)
	if result.succeeded:
		lines = result.split('\n')
	for site in env['sites']:
		start = '# start [fabric] [%s]' % site['site']
		end = '# end [fabric] [%s]' % site['site']
		try:
			# Remove the presvious section
			start_index = lines.index(start)
			end_index = lines.index(end)
			lines = lines[0:start_index] + lines[end_index+1:]
		except:
			pass
		if site.has_key('crontab') and site['crontab']:
			lines.append(start)
			for job in site['crontab']:
				if isinstance(job, (str, unicode)):
					lines.append('0 0 * * * cd %s && python manage.py %s --settings %s >/dev/null 2>&1' % ('/srv/www/%s' % env['project'], job, site['settings_module']))
				else:
					lines.append(' '.join())
			lines.append(end)
	if lines:
		import sha
		lines = lines.join('\n')
		crontab_file = '/tmp/fabric_crontab_' + sha.new(lines).digest().encode('hex')[0:10]
		sudo('touch ' + crontab_file)
		append(crontab_file, lines)
		sudo('crontab -u %s %s' % (nginx_user, crontab_file))
		sudo('rm ' + crontab_file)
	else:
		sudo('crontab -r %s' % nginx_user, warn_only=True)

@task
def deploy(raw=False):
	""" Deploys the latest version of the code. """
	__deploy_webapp()
	for site in env['sites']:
		if site.has_key('local_tests') and site['local_tests']:
			local('python manage.py test %s --settings %s' % (' '.join(site['local_tests']), site['settings_module']))
	sudo('service uwsgi stop')
	sudo('service nginx stop')
	sync()
	with own_project():
		with cd('/srv/www/%s' % env['project']):
			for site in env['sites']:
				sudo('python manage.py syncdb --settings %s --noinput' % site['settings_module'])
				sudo('python manage.py migrate --settings %s --noinput' % site['settings_module'])
				with settings(warn_only=True):
					# The collectstatic with --clear with raise an exception and fail if the static directory does not already exist, so retry without --clear if it fails
					result = sudo('python manage.py collectstatic --settings %s --noinput --clear' % site['settings_module'])
					if result.failed:
						sudo('python manage.py collectstatic --settings %s --noinput' % site['settings_module'])
				if site.has_key('remote_tests') and site['remote_tests']:
					with settings(warn_only=True):
						sudo('python manage.py test %s --settings %s ' % (' '.join(site['remote_tests']), site['settings_module']))
	sudo('service uwsgi start')
	sudo('service nginx start')
	deploy_crontab()

@task
def reboot():
	""" Reboot the server. """
	sudo('shutdown -r now')

def __uname():
	""" Print the remote uname. """
	run('uname -a')

def __ntp():
	""" Sync the time on the host using network time. """
	sudo('ntpdate pool.ntp.org')

def __exports():
	""" Prints commands to export AWS environment variables. Example use: eval `fab --hide running,status e:production x:exports` """
	for site in env['sites']:
		try:
			module = import_module(site['settings_module'])
			# Command line interface variables
			print 'export AWS_ACCESS_KEY=%s' % module.AWS_ACCESS_KEY_ID
			print 'export AWS_SECRET_KEY=%s' % module.AWS_SECRET_ACCESS_KEY
			# Boto variables
			print 'export AWS_ACCESS_KEY_ID=%s' % module.AWS_ACCESS_KEY_ID
			print 'export AWS_SECRET_ACCESS_KEY=%s' % module.AWS_SECRET_ACCESS_KEY
		except:
			pass

# Archive functions

def git_archive():
	""" Creates a git archive with all submodules included. Uses gnutar so the archive is compatible on linux. """
	if env.has_key('project'):
		project = env['project']
	else:
		project = 'project'
	local("git archive HEAD -o %s.tar; git submodule foreach 'git archive --prefix ${path}/ HEAD -o ../temp.tar; gnutar -Af ../%s.tar ../temp.tar; rm ../temp.tar'; gzip -f %s.tar" % (project, project, project))

def git_raw_archive():
	""" Creates an archive of files not based on the git repository. For testing uncommitted code. """
	if env.has_key('project'):
		project = env['project']
	else:
		project = 'project'
	local("gnutar -cf %s.tar .; gzip -f %s.tar" % (project, project))

@task
def archive(raw=False):
	""" Creates a git archive with all submodules included. Uses gnutar so the archive is compatible on linux. Optional argument: true for a raw archive. """
	if raw:
		git_raw_archive()
	else:
		git_archive()

MYSQL_CLIENT_INSTRUCTIONS = """1. Download the connector from: http://dev.mysql.com/downloads/connector/c/
2. Install the connector into /usr/local/
3. Add the correct /usr/local/mysql-connector-c-<your version here>/bin directory to PATH
4. Add the correct /usr/local/mysql-connector-c-<your version here>/lib directory to DYLD_LIBRARY_PATH"""

MYSQL_PYTHON_INSTRUCTIONS = """1. Download MySQL-python from: https://pypi.python.org/pypi/MySQL-python
2. export CFLAGS=-Wunused-command-line-argument-hard-error-in-future
3. python setup.py build
4. sudo python setup.py install
"""

PYCRYPTO_INSTRUCTIONS = """1. Download pycrypto from: https://www.dlitz.net/software/pycrypto/
2. export CFLAGS=-Wunused-command-line-argument-hard-error-in-future
3. ./configure
4. python setup.py build
5. sudo python setup.py install
"""

def __localsetup():
	""" Installs packages necessary for a local mac setup. """
	with settings(warn_only=True):
		packages = env['pip_packages'] if env.has_key('pip_packages') else django_pip_packages
		# Check to make sure mysql client libraries are installed
		if 'mysql-python' in packages:
			result = local('which mysql_config', True)
			if not result.succeeded:
				# Print mysql client setup instructions and exit
				print red('You do not have the mysql client libraries installed.', True)
				print 'Please follow these instructions first:'
				print MYSQL_CLIENT_INSTRUCTIONS
				return

		try:
			import MySQLdb
		except:
			import platform
			if platform.system() == 'Darwin':
				print red('You do not have MySQL-python installed.', True)
				print 'Please install first from easy_install, pip, or the instructions below:'
				print MYSQL_PYTHON_INSTRUCTIONS
				return

		try:
			import Crypto
		except:
			import platform
			if platform.system() == 'Darwin':
				print red('You do not have pycrypto installed.', True)
				print 'Please install first from easy_install, pip, or the instructions below:'
				print PYCRYPTO_INSTRUCTIONS
				return

		print 'The following packages will be installed:'
		print '\n'.join(packages)
		print 'Continue? y/n:'
		i = None
		while not i: i = raw_input().lower()
		if i[0] != 'y':
			return
		result = local('which pip', True)
		if result.succeeded:
			for package in packages:
				local('sudo pip install %s' % package)
				if package == 'distribute':
					local('sudo pip install --upgrade distribute')
		else:
			for package in packages:
				if package.startswith('git+git'):
					package = package.replace('git+git', 'https', 1)
					package = package.replace('.git#', '/tarball/master#', 1)
				local('sudo easy_install %s' % package)
				if package == 'distribute':
					local('sudo easy_install -U distribute')
		print ''
		print bold('Note: if you have webapps with this project, these need to be setup seperately - potentially requiring nodejs and running npm install.')

X_COMMAND_MAP = {
					'e': e,
					's': s,
					'localsetup': __localsetup,
					'setups3': __setup_buckets,
					'setupnginx': __setup_nginx,
					'setupuwsgi': __setup_uwsgi,
					'generates3ajax': __generate_s3_ajax,
					'dryrun': __sync_dry_run,
					'uname': __uname,
					'ntp': __ntp,
					'exports': __exports
				}

@task
def x(command=None):
	""" Run another internal command meant for development purposes. For a full list run: fab x """
	if not command:
		print 'Available x commands:\n'
		justification = len(max(X_COMMAND_MAP.keys(), key=len)) + 1
		for name in sorted(X_COMMAND_MAP.keys()):
			print '    %s%s' % (name.ljust(justification), X_COMMAND_MAP[name].__doc__)
	else:
		fun = X_COMMAND_MAP.get(command, None)
		if fun:
			fun()
		else:
			print red('\nError: no such command.\n', True)
			x()



import os
import re
import StringIO
import sys

from fabric.api import task, run, sudo, put, cd, env, hide, settings, local

from .shared import *
from .deploy import *
from .hooks import hook
import x, services

__all__ = ['services', 'environments', 'set_environment', 'set_default_environment', 'SiteType', 'presets', 'e', 's', 'vagrant', 'deploy', 'manage', 'setup', 'install', 'config', 'siteinstall', 'siteconfig', 'restart', 'stop', 'start', 'x']

# Python 2.7.9 enables strict SSL cert checking, so S3 buckets with dots will no longer work
# ssl.CertificateError: hostname 'my.bucket.name.s3.amazonaws.com' doesn't match either of '*.s3.amazonaws.com', 's3.amazonaws.com'
# https://www.python.org/dev/peps/pep-0476/
import ssl
try:
	_create_unverified_https_context = ssl._create_unverified_context
except AttributeError:
	# Legacy Python that doesn't verify HTTPS certificates by default
	pass
else:
	# Handle target environment that doesn't support HTTPS verification
	ssl._create_default_https_context = _create_unverified_https_context

# Add the project path to sys.path, so that utils.fab can import from a site's settings_modules
# Will work for both a fabfile directory or file
sys.path.append(os.path.dirname(env['real_fabfile'].rstrip('/')))

class CompactStdout(object):
	def __init__(self):
		self.prefix = None
		self.in_percent = False
	def isatty(self):
		return sys.__stdout__.isatty()
	def flush(self):
		sys.__stdout__.flush()
	def write(self, s):
		# Remove all the newlines and before adding them back in, effectively removing empty lines
		if not s:
			return
		s = s.strip()
		if not s:
			self.prefix = None
			return
		if self.prefix:
			contains_percent = (re.search("^\d*%", s) or re.search("\d*%$", s))
			if self.in_percent:
				if contains_percent:
					sys.__stdout__.write('\r')
				else:
					sys.__stdout__.write('\n')
			sys.__stdout__.write(blue(self.prefix))
			self.prefix = None
			self.in_percent = contains_percent
			sys.__stdout__.write(s)
			if not self.in_percent:
				sys.__stdout__.write('\n')
		elif s.startswith('[%s]' % env.host_string):
			self.prefix = s + ' '
			return
		elif s.lower() == 'done.':
			return
		else:
			sys.__stdout__.write(s)
			sys.__stdout__.write('\n')

sys.stdout = CompactStdout()
env.colorize_errors = True
env.check_production_requirements = True

def check_production_requirements():
	if env.check_production_requirements:
		# Make sure the current branch is master, otherwise warn the user
		with hide('everything'), settings(warn_only=True):
			result = local('git status | head -1', capture=True)
		if result.succeeded and result.split()[-1] != 'master':
			answer = raw_input('Warning: Not on branch master, continue anyways? (n): ')
			if not 'yes'.startswith(answer.lower()):
				exit(0)

@task
def f():
	""" Force the operation by ignoring any requirements and restrictions. """
	env.check_production_requirements = False

@task
def e(name):
	""" Set the environment for use with other commands. """
	if name:
		name = name.lower()
	if environments.has_key(name):
		set_environment(environments[name])
	else:
		print "Error: Unknown environment specified."

@task
def s(name):
	""" Set the site to use with other commands. Required: siteinstall, siteconfig, manage, and x:deploywebapp, Optional: setup, deploy. Ignored by all other tasks.  """
	if name:
		name = name.lower()
	if not env.has_key('sites'):
		print red('Error: An environment with sites must be specified before a site.')
		exit(1)
	if not env['sites'].has_key(name):
		print red('Error: Unknown site specified.')
		exit(1)
	env['site'] = env['sites'][name]

# Supported vars: http://docs.fabfile.org/en/1.4.0/usage/execution.html#ssh-config
__SSH_CONFIG_MAP = { 'User': 'user', 'Port': 'port', 'HostName': 'hosts', 'IdentityFile': 'key_filename', 'ForwardAgent': 'forward_agent' }

@task
def vagrant():
	""" Use to override the fab environment with information taken from vagrant ssh_config. """
	f()
	result = local('vagrant ssh-config', capture=True)
	if result.failed:
		print red(result)
		exit(1)
	results = result.splitlines()
	for line in results:
		key, value = line.strip().split()
		if key in __SSH_CONFIG_MAP:
			env[__SSH_CONFIG_MAP[key]] = value.strip('"')
	if type(env['hosts']) is str:
		env['hosts'] = [env['hosts']]

@task
def deploy():
	""" Deploy the project to the server using rsync, if site is specified, then will deploy a webapp. """
	check_production_requirements()
	site = env.get('site')
	with hook('deploy'):
		# Handle webapp deployments
		deploy_webapp()
		# Handle Django deployments
		sites = env['sites'].values() if site is None else [site]
		sites = [site for site in sites if site['type'] == SiteType.DJANGO]
		if sites:
			for site in sites:
				if site.has_key('local_tests') and site['local_tests']:
					local('python manage.py test %s --settings %s' % (' '.join(site['local_tests']), site['settings_module']))
			uwsgi = find_service('uwsgi')
			nginx = find_service('nginx')
			uwsgi.stop()
			nginx.stop()
			django_sync(sites)
			with own_project():
				with cd('/srv/www/%s' % env['project']):
					for site in sites:
						# Only syncdb for versions below 1.7
						#sudo('python manage.py syncdb --settings %s --noinput' % site['settings_module'])
						sudo('python manage.py migrate --settings %s --noinput' % site['settings_module'])
						with settings(warn_only=True):
							# The collectstatic with --clear with raise an exception and fail if the static directory does not already exist, so retry without --clear if it fails
							result = sudo('python manage.py collectstatic --settings %s --noinput --clear' % site['settings_module'])
							if result.failed:
								sudo('python manage.py collectstatic --settings %s --noinput' % site['settings_module'])
						if site.has_key('remote_tests') and site['remote_tests']:
							with settings(warn_only=True):
								sudo('python manage.py test %s --settings %s ' % (' '.join(site['remote_tests']), site['settings_module']))
						# Enable the site for nginx and uwsgi
						with hide('warnings'), settings(warn_only=True):
							sudo('ln -s /etc/nginx/sites-available/%s.conf /etc/nginx/sites-enabled/%s.conf' % (site['name'], site['name']))
							sudo('ln -s /etc/uwsgi/apps-available/%s.ini /etc/uwsgi/apps-enabled/%s.ini' % (site['name'], site['name']))
			uwsgi.start()
			nginx.start()

@task
def manage(*args):
	""" Run a management command using the supplied arguments. If site is not specified command will be run on all sites. """
	site = env.get('site')
	with hook('manage %s' % args[0], args[1:]):
		from services.nginx import NGINX_USER
		if site is not None:
			with cd('/srv/www/%s' % env['project']):
				sudo('python manage.py %s --settings %s' % (' '.join(args), site['settings_module']), user=NGINX_USER)
		else:
			for site in env['sites'].values():
				with cd('/srv/www/%s' % env['project']):
					sudo('python manage.py %s --settings %s' % (' '.join(args), site['settings_module']), user=NGINX_USER)

# Service tasks

@task
def setup(*service_names):
	""" Setup a server from scratch by installing and configuring services for the server and the sites. """
	check_production_requirements()
	install(*service_names)
	config(*service_names)
	siteinstall(*service_names)
	siteconfig(*service_names)

@task
def install(*service_names):
	""" Install services on the server. """
	# Install the minimum common packages first
	apt_get_update()
	apt_get_install('gcc', 'make', 'linux-headers-$(uname -r)', 'build-essential', 'libtool', 'autoconf', 'zip')
	for service in env['services']:
		if not service_names or service.name in service_names:
			service.install()

@task
def config(*service_names):
	""" Configure services on the server. """
	check_production_requirements()
	for service in env['services']:
		if not service_names or service.name in service_names:
			service.config()

@task
def siteinstall(*service_names):
	""" Install packages needed for services for one or all sites. """
	site = env.get('site')
	sites = env['sites'].values() if site is None else [site]
	# Run initial setup for each type of site
	site_types = set([s['type'] for s in sites])
	for t in site_types:
		if t == SiteType.DJANGO:
			apt_get_update()
			apt_get_install(*DJANGO_PACKAGES)
			pip_update()
			pip_install(*DJANGO_PIP_PACKAGES)
	# Install packages and each service for each site
	for s in sites:
		print teal(s['name'])
		if s.get('packages'):
			apt_get_install(*s['packages'])
		if s.get('pip_packages'):
			pip_install(*s['pip_packages'])
		for service in s.get('services', []):
			if not service_names or service.name in service_names:
				service.site_install(s)

@task
def siteconfig(*service_names):
	""" Configure the services for one or all sites. """
	check_production_requirements()
	site = env.get('site')
	sites = env['sites'].values() if site is None else [site]
	for s in sites:
		print teal(s['name'])
		for service in s.get('services', []):
			if not service_names or service.name in service_names:
				service.site_config(s)

@task
def restart(*service_names):
	""" Restart one or more services. """
	for name in service_names:
		service = find_service(name)
		if service:
			service.restart()
		else:
			print red('No such service: %s' % name)

@task
def start(*service_names):
	""" Start one or more services. """
	for name in service_names:
		service = find_service(name)
		if service:
			service.start()
		else:
			print red('No such service: %s' % name)

@task
def stop(*service_names):
	""" Stop one or more services. """
	for name in service_names:
		service = find_service(name)
		if service:
			service.stop()
		else:
			print red('No such service: %s' % name)

# x commands

X_COMMAND_MAP = {
	'e': e,
	's': s,
	'installkey': x.install_key,
	'localsetup': x.local_setup,
	'generates3ajax': x.generate_s3_ajax,
	'deploywebapp': deploy_webapp,
	'dryrun': django_sync_dry_run,
	'uname': x.uname,
	'ntp': x.ntp,
	'exports': x.exports,
	'archive': x.archive,
	'reboot': x.reboot,
	'runwebapp': x.run_webapp
}

@task
def x(command=None, *args):
	""" Run another internal command meant for development purposes. For a full list run: fab x """
	if not command:
		print 'Available x commands:\n'
		justification = len(max(X_COMMAND_MAP.keys(), key=len)) + 1
		for name in sorted(X_COMMAND_MAP.keys()):
			print '    %s%s' % (name.ljust(justification), X_COMMAND_MAP[name].__doc__)
	else:
		fun = X_COMMAND_MAP.get(command, None)
		if fun:
			fun(*args)
		else:
			print red('\nError: no such command.\n')
			x()



import os
try:
	from urllib.parse import unquote
except ImportError:
	from urllib import unquote

from django.conf import settings
from django.contrib.staticfiles import finders
from django.http import HttpResponse, HttpResponseRedirect, Http404
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from django.views.static import serve

from .decorators import require_referer
from .middleware import AssumeMiddleware

@require_referer()
def iplocation(request):
	"""Resolves the users's remote address into a location.
	
	Returns with a callback and argument similar to what is expected from navigator.geolocation.getCurrentPosition()
	"""
	from django.contrib.gis.geoip import GeoIP
	loc = None
	if request.META.has_key('REMOTE_ADDR'):
		loc = GeoIP().lat_lon(request.META['REMOTE_ADDR'])
	if loc:
		return HttpResponse("iplocationLoaded({'coords':{'latitude': %f, 'longitude': %f}});" % loc, content_type='text/javascript')
	else:
		return HttpResponse("iplocationLoaded(null);", content_type='text/javascript')

@require_POST
@csrf_exempt
def session_cross_subdomain(request):
	"""Cross a session over from one subdomain to another, where session backends are shared. Require a POST so that sessionid doesn't get logged anywhere."""
	next = request.POST.get('next', None)
	if next:
		response = HttpResponseRedirect(next)
	else:
		response = HttpResponse()
	if request.POST.has_key('sessionid'):
		response.set_cookie('sessionid', request.POST['sessionid'])
	return response

def webapp_serve(request, path, document_root=None):
	"""View for debugging (runwebapp), adds index.html to whatever path if needed. Will try to resolve static files when STATIC_URL prefix is missing."""
	index = settings.WEBAPP_INDEX if hasattr(settings, 'WEBAPP_INDEX') else 'index.html'
	if path.endswith('/'):
		return serve(request, path + index, document_root)
	else:
		# Determine if path is directory
		path = os.path.normpath(unquote(path))
		path = path.lstrip('/')
		fullpath = os.path.join(document_root, path)	
		if os.path.isdir(fullpath):
			return serve(request, path + '/' + index, document_root)
		else:
			try:
				return serve(request, path, document_root)
			except Http404 as e:
				result = finders.find(path)
				if not result:
					raise e
				else:
					result = result[0] if isinstance(result, (list, tuple)) else result
					return serve(request, path, result[:-len(path)])



import os
import urllib2

from django.conf import settings
from django.contrib.staticfiles import finders

from .colors import bold, red

# NOTE: the path must always end with a /
URL_META_DATA = 'http://169.254.169.254/latest/%s'

def get_user_data():
	"""Get the user-data associated with the current instance."""
	path = URL_META_DATA % 'user-data/'
	return urllib2.urlopen(path).read()

def upload_to_s3(bucket, directory=None, files=None, prefix=None, callback=None):
	"""Uploads files to an s3 bucket.

	Upload either an entire directory with files=None, or specific files with and optional directory prefix.
	If callback is provided it will do a callback for each file uploaded (filename, keyname)
	"""
	
	if bucket is None:
		raise ValueError('Bucket must be specified.')
	if directory is None and files is None:
		raise ValueError('Directory and/or files must be specified.')
	
	# Setup and connect boto
	from boto.s3.connection import S3Connection
	from boto.s3.bucket import Bucket
	from boto.s3.key import Key
	c = S3Connection(settings.AWS_ACCESS_KEY_ID, settings.AWS_SECRET_ACCESS_KEY)
	k = Key(Bucket(c, bucket))

	# Fix the prefix
	# prefix itself shouldn't have a / prefix itself but should end with /
	if prefix:
		prefix = prefix.lstrip('/')
		if prefix and not prefix.endswith('/'):
			prefix = prefix + '/'

	if files:
		# Upload individual files
		if directory:
			keys = [filename.lstrip('/') for filename in files]
			files = [os.path.join(directory, filename) for filename in files]
		else:
			keys = [os.path.split(filename)[1] for filename in files]
		for i, filename in enumerate(files):
			if prefix:
				k.key = prefix + keys[i]
			else:
				k.key = keys[i]
			if callback:
				callback(filename, k.key)
			k.set_contents_from_filename(filename)
	elif directory:
		# Upload an entire directory
		def __upload(arg, dirname, names):
			# arg is the starting directory
			for name in names:
				filename = os.path.join(dirname, name)
				if not os.path.isdir(filename) and not os.path.islink(filename) and not name.startswith('.'):
					key = filename[len(arg):]
					if key.startswith('/'):
						key = key[1:]
					if prefix:
						key = prefix + key
					k.key = key
					if callback:
						callback(filename, k.key)
					k.set_contents_from_filename(filename)
		os.path.walk(directory, __upload, directory)

def delete_from_s3(bucket, prefix=None, callback=None):
	"""Remove all files with the prefix specified from the bucket."""
	
	if bucket is None:
		raise ValueError('Error: Bucket must be specified.')
	
	# Setup boto
	from boto.s3.connection import S3Connection
	from boto.s3.bucket import Bucket
	from boto.s3.key import Key

	# Fix the prefix
	if prefix:
		prefix = prefix.lstrip('/')

	# Connect to S3, list the contents, and remove all of the keys
	c = S3Connection(settings.AWS_ACCESS_KEY_ID, settings.AWS_SECRET_ACCESS_KEY)
	b = Bucket(c, bucket)
	result_set = b.list(prefix=prefix)
	if callback:
		for key in result_set:
			callback(key.name)
	result = b.delete_keys([key.name for key in result_set])



import copy
import json
import os
import StringIO
import yaml

from fabric.api import cd, put, settings, sudo, env
from fabric.context_managers import shell_env
from fabric.contrib.files import exists

from .cron import add_crontab_section, remove_crontab_section, CronSchedule, CronJob
from .postgres import Postgres, POSTGRES_USER
from .service import Service
from ..hooks import hook
from ..shared import apt_get_install, pip_install, red, chown, find_service

_PACKAGE_URL = 'http://dl.bintray.com/snowplow/snowplow-generic/snowplow_emr_r77_great_auk.zip'
_MASTER_URL = 'https://codeload.github.com/snowplow/snowplow/zip/master'
_MASTER_FILE = 'snowplow-master.zip'

_INSTALL_DIR = '/opt/snowplow'
_MASTER_DIR = os.path.join(_INSTALL_DIR, 'snowplow-master')
_RUNNER_PATH = os.path.join(_INSTALL_DIR, 'snowplow-emr-etl-runner')
_LOADER_PATH = os.path.join(_INSTALL_DIR, 'snowplow-storage-loader')
_CONFIG_PATH = os.path.join(_INSTALL_DIR, 'config.yml')
_RESOLVER_PATH = os.path.join(_INSTALL_DIR, 'iglu_resolver.json')
_CREATE_TABLE_PATH = os.path.join(_MASTER_DIR, '4-storage/postgres-storage/sql/atomic-def.sql')
_ENRICHMENTS_PATH = os.path.join(_MASTER_DIR, '3-enrich/config/enrichments')
_DEFAULT_RESOLVER = {
	'schema': 'iglu:com.snowplowanalytics.iglu/resolver-config/jsonschema/1-0-0',
	'data': {
		'cacheSize': 500,
		'repositories': [
			{
				'name': 'Iglu Central',
				'priority': 0,
				'vendorPrefixes': [ 'com.snowplowanalytics' ],
				'connection': { 'http': { 'uri': 'http://iglucentral.com' } }
			}
		]
	}
}
_CRONTAB_USER = 'root'
_CRONTAB_SECTION = '[snowplow]'
_RUNNER_COMMAND = ' '.join((_RUNNER_PATH, '--config', _CONFIG_PATH, '--resolver', _RESOLVER_PATH))
_LOADER_COMMAND = ' '.join((_LOADER_PATH, '--config', _CONFIG_PATH))

_DEFAULT_SETTINGS = {
	'schedule': CronSchedule(45, 2),
	'chain': None,
	'enrichments': _ENRICHMENTS_PATH,
	'runner_skip': (),
	'loader_skip': ()
}

def _config_postgres(target):
	# Assumes that the target database is already installed, running, and setup with the correct credentials but will try to create both the database and table
	apt_get_install('postgresql-client')
	pg_env = {
		'PGHOST': target['host'],
		'PGPORT': str(target.get('port', '5432')),
		'PGUSER': target['username'],
		'PGPASSWORD': target['password'],
		'PGDATABASE': target['database']
	}
	# Create the user if a local database
	if find_service(Postgres.name) is not None:
		with settings(warn_only=True):
			sudo('createuser --echo --createdb --no-superuser --no-createrole %s' % target['username'], user=POSTGRES_USER)
			sudo("psql --echo-queries -c \"ALTER USER %s WITH PASSWORD '%s';\"" % (target['username'], target['password']), user=POSTGRES_USER)
	with shell_env(**pg_env):
		with settings(warn_only=True):
			# Create the database
			sudo('createdb --echo %s' % target['database'])
			# Create the table - currently only atomic.events is supported as the table name
			if target.get('table', 'atomic.events') != 'atomic.events':
				print red('Only atomic.events is supported as a snowplow postgres storage table name.')
				return
			sudo('psql -f %s' % _CREATE_TABLE_PATH)

class Snowplow(Service):
	name = 'snowplow'
	script = None

	def install(self):
		with hook('install %s' % self.name, self):
			if not exists(_INSTALL_DIR):
				apt_get_install('default-jre', 'unzip')
				sudo('mkdir %s' % _INSTALL_DIR)
				with cd(_INSTALL_DIR):
					sudo('wget --no-clobber %s' % _PACKAGE_URL)
					sudo('unzip %s' % _PACKAGE_URL.split('/')[-1])
					sudo('wget --no-clobber --output-document %s %s' % (_MASTER_FILE, _MASTER_URL))
					sudo('unzip %s' % _MASTER_FILE)

	def config(self):
		# Possible configuration options are custom repositories by setting the repositories setting to an array of repository objects
		with hook('config %s' % self.name, self):
			resolver = copy.deepcopy(_DEFAULT_RESOLVER)
			repositories = self.settings.get('repositories')
			if repositories:
				resolver['data']['repositories'].extend(repositories)
			chown(put(StringIO.StringIO(json.dumps(resolver, indent=4)), _RESOLVER_PATH, use_sudo=True, mode=0644))
			chown(put(self.settings['config_file'], _CONFIG_PATH, use_sudo=True, mode=0644))

			# Read the config file for storage configuration
			with open(self.settings['config_file']) as f:
				config = yaml.load(f)
				if config.get('storage') and config['storage'].get('targets'):
					for target in config['storage']['targets']:
						if target.get('type') == 'postgres':
							_config_postgres(target)

			# Schedule cron jobs
			chain = self.settings.get('chain', _DEFAULT_SETTINGS['chain'])
			if not chain.log_name:
				chain.log_name = self.name
			loader_skip = ','.join(self.settings.get('loader_skip', _DEFAULT_SETTINGS['loader_skip']))
			if loader_skip:
				loader_skip = '--skip ' + loader_skip
			loader_job = CronJob(_LOADER_COMMAND + loader_skip, log_name=self.name, chain=chain)
			runner_skip = ','.join(self.settings.get('runner_skip', _DEFAULT_SETTINGS['runner_skip']))
			if runner_skip:
				runner_skip = '--skip ' + runner_skip
			runner_enrichments = self.settings.get('enrichments', _DEFAULT_SETTINGS['enrichments'])
			if runner_enrichments:
				runner_enrichments = '--enrichments ' + runner_enrichments
			runner_job = CronJob(_RUNNER_COMMAND + runner_enrichments + runner_skip, log_name=self.name, schedule=self.settings.get('schedule', _DEFAULT_SETTINGS['schedule']), chain=loader_job)

			# Configure cron with the first site with the Snowplow service or just the active or first site if no Snowplow service is found
			sites = env['sites'].values() or []
			snowplow_site = env.get('site') or sites[0]
			for site in sites:
				if site:
					found = False
					for service in site.get('services', []):
						if isinstance(service, Snowplow):
							found = True
							break
					if found:
						snowplow_site = site
						break
			remove_crontab_section(_CRONTAB_USER, _CRONTAB_SECTION)
			add_crontab_section(_CRONTAB_USER, _CRONTAB_SECTION, runner_job, snowplow_site)




def add_widget_attrs(form, filter=None, **attrs):
	"""Add attributes to all widgets in a form before rendering. Sometimes easier than specifying in the form."""
	for attr in attrs:
		for field in form.fields.values():
			if filter and not filter(field):
				continue
			if field.widget.attrs.has_key(attr):
				field.widget.attrs[attr] = '%s %s' % (field.widget.attrs[attr], attrs[attr])
			else:
				field.widget.attrs[attr] = attrs[attr]
